#!/usr/bin/env python3
# ePOS file format reader used by atom probe microscopists.
#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

# pylint: disable=no-member,duplicate-code

import os

import numpy as np

from ifes_apt_tc_data_modeling.nexus.nx_field import NxField

from ifes_apt_tc_data_modeling.utils.mmapped_io import get_memory_mapped_data


class ReadEposFileFormat():
    """Read *.epos file format."""

    def __init__(self, filename: str):
        assert len(filename) > 5, "ePOS file incorrect filename ending!"
        assert filename.lower().endswith(".epos"), \
            "ePOS file incorrect file type!"
        self.filename = filename

        self.filesize = os.path.getsize(self.filename)
        assert self.filesize % 11 * 4 == 0, \
            "ePOS filesize not integer multiple of 11*4B!"
        assert np.uint32(self.filesize / (11 * 4)) < np.iinfo(np.uint32).max, \
            "ePOS file is too large, currently only 2*32 supported!"
        self.number_of_events = np.uint32(self.filesize / (11 * 4))

        # https://doi.org/10.1007/978-1-4614-3436-8 for file format details
        # dtyp_names = ["Reconstructed position along the x-axis (nm)",
        #               "Reconstructed position along the y-axis (nm)",
        #               "Reconstructed position along the z-axis (nm)",
        #               "Reconstructed mass-to-charge-state ratio (Da)",
        #               "Raw time-of-flight (ns)",
        #               "Standing voltage (V)",
        #               "Pulsed voltage (V)",
        #               "Ion impact x-coordinate at the detector (mm)",
        #               "Ion impact y-coordinate at the detector (mm)",
        #               "Number of pulses since the last detected ion (pulses)",
        #               "Hit multiplicity (ions)"]
        # raw = np.fromfile( fnm, dtype= {"names": dtyp_names,
        # "formats": (, ">f4",">f4",">f4",">f4",">f4",">f4",">u4",">u4") } )

    def get_reconstructed_positions(self):
        """Read xyz columns."""

        xyz = NxField()
        xyz.typed_value = np.zeros(
            [self.number_of_events, 3], np.float32)
        xyz.unit = "nm"

        xyz.typed_value[:, 0] = \
            get_memory_mapped_data(self.filename, ">f4",
                                   0 * 4, 11 * 4, self.number_of_events)  # x
        xyz.typed_value[:, 1] = \
            get_memory_mapped_data(self.filename, ">f4",
                                   1 * 4, 11 * 4, self.number_of_events)  # y
        xyz.typed_value[:, 2] = \
            get_memory_mapped_data(self.filename, ">f4",
                                   2 * 4, 11 * 4, self.number_of_events)  # z
        return xyz

    def get_mass_to_charge_state_ratio(self):
        """Read mass-to-charge-state-ratio column."""
        m_n = NxField()
        m_n.typed_value = np.zeros(
            [self.number_of_events, 1], np.float32)
        m_n.unit = "Da"

        m_n.typed_value[:, 0] = \
            get_memory_mapped_data(self.filename, ">f4",
                                   3 * 4, 11 * 4, self.number_of_events)
        return m_n

    def get_raw_time_of_flight(self):
        """Read raw (uncorrected) time-of-flight."""
        raw_tof = NxField()
        raw_tof.typed_value = np.zeros(
            [self.number_of_events, 1], np.float32)
        raw_tof.unit = "ns"

        # according to DOI: 10.1007/978-1-4899-7430-3 raw time-of-flight
        # i.e. this is an uncorrected time-of-flight
        # for which effects uncorrect?
        # Only the proprietary IVAS/APSuite source code knows for sure
        raw_tof.typed_value[:, 0] = \
            get_memory_mapped_data(self.filename, ">f4",
                                   4 * 4, 11 * 4, self.number_of_events)
        return raw_tof

    def get_standing_voltage(self):
        """Read standing voltage."""
        # according to DOI: 10.1007/978-1-4899-7430-3
        # standing voltage on the specimen
        # according to DOI: 10.1007/978-1-4614-8721-0 also-known as DC voltage
        dc_voltage = NxField()
        dc_voltage.typed_value = np.zeros(
            [self.number_of_events, 1], np.float32)
        dc_voltage.unit = "kV"
        # different to the above-mentioned references Gault et al. state
        # that standing and pulse_voltage are in V instead of kV

        dc_voltage.typed_value[:, 0] = \
            get_memory_mapped_data(self.filename, ">f4",
                                   5 * 4, 11 * 4, self.number_of_events)
        return dc_voltage

    def get_pulse_voltage(self):
        """Read pulse voltage."""
        # according to DOI: 10.1007/978-1-4899-7430-3
        # additional voltage to trigger field evaporation in case
        # of high-voltage pulsing, 0 for laser pulsing
        pu_voltage = NxField()
        pu_voltage.typed_value = np.zeros(
            [self.number_of_events, 1], np.float32)
        pu_voltage.unit = "kV"

        pu_voltage.typed_value[:, 0] = \
            get_memory_mapped_data(self.filename, ">f4",
                                   6 * 4, 11 * 4, self.number_of_events)
        return pu_voltage

    def get_hit_positions(self):
        """Read ion impact positions on detector."""
        hit_positions = NxField()
        hit_positions.typed_value = np.zeros(
            [self.number_of_events, 2], np.float32)
        hit_positions.unit = "mm"

        hit_positions.typed_value[:, 0] = \
            get_memory_mapped_data(self.filename, ">f4",
                                   7 * 4, 11 * 4, self.number_of_events)  # x
        hit_positions.typed_value[:, 1] = \
            get_memory_mapped_data(self.filename, ">f4",
                                   8 * 4, 11 * 4, self.number_of_events)  # y
        return hit_positions

    def get_number_of_pulses(self):
        """Read number of pulses."""
        # according to DOI: 10.1007/978-1-4899-7430-3
        # number of pulses since last event detected
        # 0 after the first ion per pulse
        # also known as $\Delta Pulse$
        npulses = NxField()
        npulses.typed_value = np.zeros(
            [self.number_of_events, 1], np.uint32)
        npulses.unit = ""

        npulses.typed_value[:, 0] = \
            get_memory_mapped_data(self.filename, ">u4",
                                   9 * 4, 11 * 4, self.number_of_events)
        return npulses

    def get_ions_per_pulse(self):
        """Read ions per pulse."""
        # according to DOI: 10.1007/978-1-4899-7430-3
        # ions per pulse, 0 after the first ion
        ions_per_pulse = NxField()
        ions_per_pulse.typed_value = np.zeros(
            [self.number_of_events, 1], np.uint32)
        ions_per_pulse.unit = ""

        ions_per_pulse.typed_value[:, 0] = \
            get_memory_mapped_data(self.filename, ">u4",
                                   10 * 4, 11 * 4, self.number_of_events)
        return ions_per_pulse



# RRNG range file reader used by atom probe microscopists.
#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

# pylint: disable=no-member,duplicate-code

import re

import numpy as np

from ifes_apt_tc_data_modeling.nexus.nx_ion import NxField, NxIon
from ifes_apt_tc_data_modeling.utils.utils import \
    create_isotope_vector, is_range_significant
from ifes_apt_tc_data_modeling.utils.definitions import MQ_EPSILON
from ifes_apt_tc_data_modeling.utils.molecular_ions import MolecularIonBuilder
from ifes_apt_tc_data_modeling.utils.molecular_ions import \
    PRACTICAL_ABUNDANCE, PRACTICAL_ABUNDANCE_PRODUCT, \
    PRACTICAL_MIN_HALF_LIFE, VERBOSE, SACRIFICE_ISOTOPIC_UNIQUENESS

from ase.data import atomic_numbers, atomic_masses, chemical_symbols


def evaluate_rrng_range_line(i: int, line: str) -> dict:
    """Evaluate information content of a single range line."""
    # example line "Range7 = 42.8160 43.3110 vol:0.04543
    #     Al:1 O:1 Name: AlO Color:00FFFF"
    # according to DOI: 10.1007/978-1-4614-8721-0
    # mqmin, mqmax, vol, ion composition is required,
    #     name and color fields are optional
    info: dict = {}
    info["identifier"] = "Range" + str(i)
    info["range"] = np.asarray([0., MQ_EPSILON], np.float64)
    info["atoms"] = []
    info["volume"] = np.float64(0.)
    info["color"] = ""
    info["name"] = ""

    tmp = re.split(r"[\s=]+", line)
    assert len(tmp) >= 6, "Line " + line + \
        " does not contain all required fields!"
    assert tmp[0] == "Range" + str(i), "Line " + line + \
        " has inconsistent line prefix!"

    assert is_range_significant(np.float64(tmp[1]), np.float64(tmp[2])), \
        "Line " + line + " insignificant range!"
    info["range"] = np.asarray([tmp[1], tmp[2]], np.float64)

    if tmp[3].lower().startswith("vol:"):
        info["volume"] = np.float64(re.split(r":", tmp[3])[1])
    if (tmp[-1].lower().startswith("color:")) and \
        (len(re.split(r":", tmp[-1])[1]) == 6):
        info["color"] = "#" + re.split(r":", tmp[-1])[1]
    # HEX_COLOR_REGEX = r"^([A-Fa-f0-9]{6}|[A-Fa-f0-9]{3})$"
    # replace r"^#( ...
    # regexp = re.compile(HEX_COLOR_REGEX)
    # if regexp.search(tmp[-1].split(r":")):

    for information in tmp[4:-1]:
        element_multiplicity = re.split(r":+", information)
        assert len(element_multiplicity) == 2, \
           "Element multiplicity is incorrectly formatted!"
        # skip vol, name, and color information
        if element_multiplicity[0].lower() == "name":
            info["name"] = str(element_multiplicity[1])
            # this captures properly formatted ranges with keyword
            # name whose name value is then however not a chemical
            # symbol but some user-defined string
            # this is a common habit to define custom names
        elif element_multiplicity[0].lower() not in ["vol", "color"]:
            # pick up what is an element name
            symbol = element_multiplicity[0]
            assert (symbol in chemical_symbols) and (symbol != "X"), \
                "Line " + line + " contains an invalid chemical symbol!"
            assert np.uint32(element_multiplicity[1]) > 0, \
                "Line " + line + " zero or negative multiplicity!"
            assert np.uint32(element_multiplicity[1]) < 256, \
                "Line " + line + " unsupport high multiplicity!"
            info["atoms"] = np.append(
                info["atoms"], [symbol] * int(element_multiplicity[1]))
    return info


class ReadRrngFileFormat():
    """Read *.rrng file format."""

    def __init__(self, filename: str):
        assert len(filename) > 5, "RRNG file incorrect filename ending!"
        assert filename.lower().endswith(".rrng"), \
            "RRNG file incorrect file type!"
        self.filename = filename

        self.rrng: dict = {}
        self.rrng["ionnames"] = []
        self.rrng["ranges"] = {}
        self.rrng["ions"] = {}
        self.rrng["molecular_ions"] = []
        self.read_rrng()

    def read_rrng(self):
        """Read content of an RRNG range file."""
        with open(self.filename, mode="r", encoding="utf8") as rrngf:
            txt = rrngf.read()

        txt = txt.replace("\r\n", "\n")  # windows to unix EOL conversion
        txt = txt.replace(",", ".")  # use decimal dots instead of comma
        txt_stripped = [line for line in txt.split("\n")
                        if line.strip() != "" and line.startswith("#") is False]
        del txt

        # see DOI: 10.1007/978-1-4899-7430-3 for further details to this
        # AMETEK/Cameca"s *.rrng file format

        # first, parse [Ions] section, which holds a list of element names
        # there are documented cases where experimentalists add custom strings
        # to specify ranges they consider special
        # these are loaded as user types
        # with isotope_vector np.iinfo(np.uint16).max
        where = [idx for idx, element in
                 enumerate(txt_stripped) if element == "[Ions]"]
        assert isinstance(where, list), "Section [Ions] not found!"
        assert len(where) == 1, "Section [Ions] not found or ambiguous!"
        current_line_id = where[0] + 1

        tmp = re.split(r"[\s=]+", txt_stripped[current_line_id])
        assert len(tmp) == 2, "[Ions]/Number line corrupted!"
        assert tmp[0] == "Number", "[Ions]/Number incorrectly formatted!"
        assert tmp[1].isnumeric(), "[Ions]/Number not a number!"
        number_of_ion_names = int(tmp[1])
        assert number_of_ion_names > 0, "No ion names defined!"
        current_line_id += 1
        for i in np.arange(0, number_of_ion_names):
            tmp = re.split(r"[\s=]+", txt_stripped[current_line_id + i])
            assert len(tmp) == 2, "[Ions]/Ion line corrupted!"
            assert tmp[0] == "Ion" + str(i + 1), \
                "[Ions]/Ion incorrectly formatted!"
            assert isinstance(tmp[1], str), "[Ions]/Name not a string!"
            self.rrng["ionnames"].append(tmp[1])  # [tmp[0]] = tmp[1]

        # second, parse [Ranges] section
        where = [idx for idx, element in
                 enumerate(txt_stripped) if element == "[Ranges]"]
        assert isinstance(where, list), "Section [Ranges] not found!"
        assert len(where) == 1, "Section [Ranges] not found or ambiguous!"
        current_line_id = where[0] + 1

        tmp = re.split(r"[\s=]+", txt_stripped[current_line_id])
        assert len(tmp) == 2, "[Ranges]/Number line corrupted!"
        assert tmp[0] == "Number", "[Ranges]/Number incorrectly formatted!"
        assert tmp[1].isnumeric(), "[Ranges]/Number not a number!"
        number_of_ranges = int(tmp[1])
        assert number_of_ranges > 0, "No ranges defined!"
        current_line_id += 1

        for i in np.arange(0, number_of_ranges):
            dct = evaluate_rrng_range_line(i + 1, txt_stripped[current_line_id + i])
            assert dct, \
                "Line " + txt_stripped[current_line_id + i] + " is corrupted!"

            m_ion = NxIon(isotope_vector=create_isotope_vector(
                dct["atoms"]), charge_state=0)
            m_ion.add_range(dct["range"][0], dct["range"][1])
            m_ion.comment = NxField(dct["name"], "")
            m_ion.color = NxField(dct["color"], "")
            m_ion.volume = NxField(dct["volume"], "")
            # m_ion.report()

            crawler = MolecularIonBuilder(
                min_abundance=PRACTICAL_ABUNDANCE,
                min_abundance_product=PRACTICAL_ABUNDANCE_PRODUCT,
                min_half_life=PRACTICAL_MIN_HALF_LIFE,
                sacrifice_uniqueness=SACRIFICE_ISOTOPIC_UNIQUENESS,
                verbose=VERBOSE)
            recovered_charge_state, m_ion_candidates = crawler.combinatorics(
                m_ion.isotope_vector.typed_value,
                m_ion.ranges.typed_value[0, 0],
                m_ion.ranges.typed_value[0, 1])
            # print(f"{recovered_charge_state}")
            m_ion.charge_state = NxField(np.int8(recovered_charge_state), "")
            m_ion.update_human_readable_name()
            m_ion.add_charge_state_model(
                {"min_abundance": PRACTICAL_ABUNDANCE,
                 "min_abundance_product": PRACTICAL_ABUNDANCE_PRODUCT,
                 "min_half_life": PRACTICAL_MIN_HALF_LIFE,
                 "sacrifice_isotopic_uniqueness": SACRIFICE_ISOTOPIC_UNIQUENESS},
                m_ion_candidates)

            self.rrng["molecular_ions"].append(m_ion)
        print(self.filename + " parsed successfully")

if __name__ == "main":
    pass
    # testing
    # parsedFile = ReadRrngFileFormat("../../R31_06365-v02.rrng")



# Utility tool to work with molecular ions in atom probe microscopy.
#
# Also convenience functions are included which translate human-readable ion
# names into the isotope_vector description proposed by Kuehbach et al. in
# DOI: 10.1017/S1431927621012241 to the human-readable ion names which are use
# in P. Felfer et al.'s atom probe toolbox
#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

# pylint: disable=no-member,duplicate-code

import typing

from typing import Tuple

import numpy as np

import radioactivedecay as rd

import ase
from ase.data import atomic_numbers, atomic_masses, atomic_names

from ifes_apt_tc_data_modeling.utils.utils import \
    isotope_to_hash, hash_to_isotope, isotope_vector_to_dict_keyword

from ifes_apt_tc_data_modeling.utils.nist_isotope_data import isotopes

# do not consider isotopes with a very low natural abundance
PRACTICAL_ABUNDANCE = 0.  # 1.0e-6
# do not consider candidate isotopically different molecular ions
# if their natural abundance product is too low
PRACTICAL_ABUNDANCE_PRODUCT = 0.  # 1.0e-12
# do consider too shortliving isotopes
PRACTICAL_MIN_HALF_LIFE = np.inf
# many examples of ranges are not constrainted strongly enough so that
# there are many isotopes (many of which admittedly hypothetical) ones
# which are within the range, this option lifts the constraint that
# there should be only one set of isotopically different molecular ions
# and if these have all these same charge it is assumed this is the
# charge of the molecular ion
# strictly speaking however one would have to rule out every possible
# molecular ion candidate (which is non-trivial and maybe not even
# with first principles theory possible...
SACRIFICE_ISOTOPIC_UNIQUENESS = True
VERBOSE=False


class MolecularIonCandidate:
    def __init__(self, ivec=[], charge_state=0, mass_sum=0., nat_abun_prod=0., min_half_life=np.inf):
        self.isotope_vector = np.asarray(np.atleast_1d(ivec), np.uint16)
        self.charge_state = np.int8(charge_state)
        self.mass = np.float64(mass_sum)
        self.abundance_product = np.float64(nat_abun_prod)
        self.shortest_half_life = np.float64(min_half_life)

    def unique_keyword(self):
        keyword = ""
        keyword += isotope_vector_to_dict_keyword(
            np.sort(np.asarray(self.isotope_vector, np.uint16), kind="stable")[::-1])
        keyword += "__" + str(self.charge_state)
        return keyword


class MolecularIonBuilder:
    def __init__(self,
                 min_abundance=1.0e-6,
                 min_abundance_product=1.0e-6,
                 min_half_life=np.inf,
                 sacrifice_uniqueness=True,
                 verbose=False):
        self.nuclids = np.asarray([], np.uint16)
        self.element_isotopes = {}
        self.nuclid_mass = {}
        self.nuclid_abundance = {}
        self.nuclid_stable = {}  # observationally stable
        self.nuclid_unclear = {}  # unclear halflife
        self.nuclid_halflife = {}
        self.candidates = []
        self.parms = {}
        self.parms["min_abundance"] = min_abundance
        self.parms["min_abundance_product"] = min_abundance_product
        self.parms["min_half_life"] = min_half_life
        self.parms["sacrifice_isotopic_uniqueness"] = sacrifice_uniqueness
        self.parms["verbose"] = verbose

        for symbol, atomic_number in atomic_numbers.items():
            if symbol != "X":
                # assume that data from ase take preference
                # if half-life data are available in rd take these if not mark
                # the isotope that it unclear_half_life == True
                element_isotopes = []
                for mass_number in isotopes[atomic_number]:
                    half_life = np.inf
                    observationally_stable = False
                    unclear_half_life = False

                    # test if half-life data available
                    trial_nuclid_name = symbol + "-" + str(mass_number)
                    try:
                        tmp = rd.Nuclide(trial_nuclid_name)
                    except:
                        tmp = None
                    if tmp is not None:
                        half_life = tmp.half_life()
                        if np.isinf(half_life):
                            observationally_stable = True
                            # these ions are always taken as they
                            # are most relevant for practical
                            # atom probe experiments
                        else:
                            if half_life < self.parms["min_half_life"]:
                                # ignore practically short living ions
                                continue
                    else:
                        continue
                        # do not consider exotic isotopes with unclear
                        # half-life as they are likely anyway irrelevant
                        # for practical atom probe experiments
                        half_life = np.nan
                        unclear_half_life = True

                    # get ase abundance data
                    n_protons = atomic_numbers[symbol]
                    n_neutrons = mass_number - n_protons
                    mass = isotopes[atomic_numbers[symbol]][mass_number]["mass"]
                    abundance = isotopes[atomic_numbers[symbol]][mass_number]["composition"]
                    # self.nuclids.append(
                    #     Isotope(mass,
                    #             abundance,
                    #             n_protons, n_neutrons,
                    #             observationally_stable,
                    #             unclear_half_life))
                    hashvalue = isotope_to_hash(int(n_protons), int(n_neutrons))
                    self.nuclids = np.asarray(np.append(self.nuclids, hashvalue), np.uint16)
                    self.nuclid_mass[hashvalue] = np.float64(mass)
                    self.nuclid_abundance[hashvalue] = np.float64(abundance)
                    self.nuclid_stable[hashvalue] = observationally_stable
                    self.nuclid_unclear[hashvalue] = unclear_half_life
                    self.nuclid_halflife[hashvalue] = half_life
                    assert np.uint16(0) not in self.nuclid_mass.keys(), \
                        "0 must not be a key in nuclid_mass dict!"
                    element_isotopes = np.append(element_isotopes, hashvalue)
                self.element_isotopes[atomic_number] = np.sort(
                    np.asarray(element_isotopes, np.uint16), kind="stable")[::-1]
        self.nuclids = np.sort(self.nuclids, kind="stable")[::-1]
        if self.parms["verbose"] is True:
            print(f"MolecularIonBuilder initialized with {len(self.nuclids)}")

    def get_element_isotopes(self, hashvalue):
        """List of hashvalues all isotopes of element specified by hashvalue."""
        # nuclids = np.asarray([], np.uint16)
        target_p, target_n = hash_to_isotope(hashvalue)
        return self.element_isotopes[target_p]
        # for source_hashvalue in self.nuclids:
        #     # naive search for now
        #     source_p, source_n = hash_to_isotope(source_hashvalue)
        #     if source_p != target_p:
        #         continue
        #     else:
        #         nuclids = np.append(nuclids, source_hashvalue)
        # return nuclids

    def get_isotope_mass_sum(self, nuclid_arr=[]):
        """Evaluate cumulated atomic_mass of isotopes in ivec."""
        # assuming no relativistic effects or other quantum effects
        # mass loss due to charge_state considered insignificant
        mass = 0.
        for hashvalue in nuclid_arr:
            if hashvalue != 0:
                mass += self.nuclid_mass[hashvalue]
            # else:  # because ivec are systematically filled sorted in descending order !
            #   break
        return mass

    def get_natural_abundance_product(self, nuclid_arr=[]):
        abun_prod = 1.
        for hashvalue in nuclid_arr:
            if hashvalue != 0:
                abun_prod *= self.nuclid_abundance[hashvalue]
        return abun_prod

    def get_shortest_half_life(self, nuclid_arr=[]):
        min_half_life = self.parms["min_half_life"]
        for hashvalue in nuclid_arr:
            if hashvalue != 0:
                if self.nuclid_halflife[hashvalue] != np.nan:
                    min_half_life = np.min((min_half_life, self.nuclid_halflife[hashvalue]))
                    # if the min_half_life is np.inf than we know that every nuclid in the
                    # molecular ion is observationally stable
                    # if not we know that considering this molecular ion might not be a good
                    # idea or only necessary in very few cases because even if we were
                    # to find such a combination of nuclids at least one would decay in observable
                    # time and this might possibly hint that studying this molecular is tricky
                else:
                    # if we do not information about the half-life it is very likely that
                    # this is an exotic nuclids likely never found in the wild
                    return np.nan
        return min_half_life

    def combinatorics(self, element_arr, low, high):
        # RNG/RRNG range files do store element information for each range
        # BUT not isotope information, correspondingly this can yield
        # only an isotope_vector whose hashvalues have ALL in common
        # that the number of neutrons is 0, i.e. their hashvalue is the atomic_number
        # element_arr is an isotope_vector/ivec with such hashvalues
        # as an example the molecular ion C:2 H:1 will map to 6, 6, 1
        max_depth = 0  # number of non-zero entries in element_arr
        for hashvalue in element_arr:
            n_protons, n_neutrons = hash_to_isotope(hashvalue)
            if hashvalue != 0:
                max_depth += 1
        if self.parms["verbose"] is True:
            print(f"Maximum recursion depth {max_depth}")
        self.candidates = []
        if self.parms["verbose"] is True:
            print(element_arr)

        if max_depth > 0:
            depth = 0
            ith_nuclids = self.get_element_isotopes(element_arr[depth])
            cand_arr_curr = []  # combinatorially add nuclids while recursing deeper
            self.iterate_molecular_ion(
                element_arr, ith_nuclids, cand_arr_curr,
                depth, max_depth, low, high)
            if self.parms["verbose"] is True:
                print(f"Found {len(self.candidates)} candidates!")
            return self.try_to_reduce_to_unique_solution()
            # will return a tuple of charge_state and list of relevant_candidates
        else:
            return (0, [])

    def iterate_molecular_ion(self,
                              element_arr, jth_nuclids, cand_arr_prev,
                              i, max_n, low, high):
        if i < (max_n - 1):
            for nuclid in jth_nuclids:

                ixxth_nuclids = self.get_element_isotopes(element_arr[i + 1])
                cand_arr_curr = np.append(cand_arr_prev, nuclid)
                self.iterate_molecular_ion(
                    element_arr, ixxth_nuclids, cand_arr_curr,
                    i + 1, max_n, low, high)
        elif i == (max_n - 1):
            for nuclid in jth_nuclids:
                cand_arr_curr = np.append(cand_arr_prev, nuclid)
                # by this design the ivec does not necessarily remain ordered

                new_mass = self.get_isotope_mass_sum(cand_arr_curr)
                new_abun_prod = self.get_natural_abundance_product(cand_arr_curr)
                new_halflife = self.get_shortest_half_life(cand_arr_curr)

                for new_chrg in np.arange(1, 8):
                    mass_to_charge = new_mass / new_chrg;
                    if mass_to_charge < low:
                        break
                    # we can break the entire charge state generation here already instead of continue
                    # because already the current mq is left out of interval [mqmin, mqmax]
                    # and all mq in the next iterations will result in even lower mass-to-charge
                    # you walk to the left increasing your distance to the left bound
                    if mass_to_charge > high:
                        # can be optimized and broken out of earlier if testing first chrg == 1
                        # and then chrg == APTMOLECULARION_MAX_CHANGE
                        continue;
                        # must not be break here because with adding more charge
                        # we usually walk from right to left eventually into [low, high] !
                    # molecular ion is within user-specified bounds
                    self.candidates.append(
                        MolecularIonCandidate(cand_arr_curr, new_chrg, new_mass, new_abun_prod, new_halflife))
                # do not start a new recursion
        else:
            # break the recursion
            return

    def try_to_reduce_to_unique_solution(self):
        """Hueristics to identify if current candidates are unique."""
        if self.parms["verbose"] is True:
            print(f"Reduce set of {len(self.candidates)} candidates to a unique...")
        self.relevant = {}
        for cand in self.candidates:
            if cand.abundance_product >= self.parms["min_abundance_product"]:
                if np.isnan(cand.shortest_half_life) == False:
                    if cand.shortest_half_life >= self.parms["min_half_life"]:
                        keyword = cand.unique_keyword()
                        if keyword not in self.relevant.keys():
                            self.relevant[keyword] = cand
        relevant_candidates = []
        if self.parms["verbose"] is True:
            print(f"Reduced set to {len(self.relevant.keys())} relevant candidates...")
            for key, val in self.relevant.items():
                print(key)
        for key, obj in self.relevant.items():
            relevant_candidates.append(obj)
        # print(type(relevant_candidates))

        if len(self.relevant) == 0:
            if self.parms["verbose"] is True:
                print("WARNING::No relevant candidate meets all criteria!")
                print("WARNING::No solution possible for given criteria!")
            return (0, relevant_candidates)
        elif len(self.relevant) == 1:
            if self.parms["verbose"] is True:
                print("One relevant candidate which meets all criteria")
            keywords = []
            for key in self.relevant.keys():
                if isinstance(key, str):
                    keywords.append(key)
            assert len(keywords) >= 1, "List of relevant keywords is empty!"
            return (self.relevant[keywords[0]].charge_state, relevant_candidates)
        else:
            if self.parms["verbose"] is True:
                print("Multiple relevant candidates meet all selection criteria")
            keywords = []
            for key in self.relevant.keys():
                if isinstance(key, str):
                    keywords.append(key)
            assert len(keywords) >= 1, "List of relevant keywords is empty!"
            charge_state = self.relevant[keywords[0]].charge_state
            for key, val in self.relevant.items():
                if val.charge_state == charge_state:
                    continue
                else:
                    if self.parms["verbose"] is True:
                        print("WARNING::Multiple relevant candidates differ in charge_state!")
                        print("WARNING::No unique solution possible for given criteria!")
                    return (0, relevant_candidates)
            # not returned yet, so all relevant candidates have the same charge_state
            if self.parms["verbose"] is True:
                print(f"Multiple relevant candidates have all the same charge_state {charge_state}")
            if self.parms["sacrifice_isotopic_uniqueness"] is True:
                return (charge_state, relevant_candidates)

            if self.parms["verbose"] is True:
                print("WARNING::Multiple relevant candidates differ in isotopes!")
                print("WARNING::But these have the same charge_state!")
                print("WARNING::No unique solution possible for given criteria!")
            return (0, relevant_candidates)



# Generic definitions when parsing molecular ions in atom probe microscopy.
#
# Also convenience functions are included which translate human-readable ion
# names into the isotope_vector description proposed by Kuehbach et al. in
# DOI: 10.1017/S1431927621012241 to the human-readable ion names which are use
# in P. Felfer et al.'s atom probe toolbox
#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

# pylint: disable=no-member,duplicate-code

import numpy as np


# restrict the number distinguished ion types
MAX_NUMBER_OF_ION_SPECIES = 256
# restrict number of atoms for molecular ion fragments
MAX_NUMBER_OF_ATOMS_PER_ION = 32
# practical and required minimum mass-resolution Da or atomic mass unit (amu)
MQ_EPSILON = np.float64(1. / 2000.)



# AMETEK APT(6) data exchange file reader used by atom probe microscopists.
#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

# pylint: disable=no-member,duplicate-code

import os

import numpy as np

import pandas as pd

from ifes_apt_tc_data_modeling.apt.apt6_utils import np_uint16_to_string

from ifes_apt_tc_data_modeling.apt.apt6_headers import AptFileHeaderMetadata

from ifes_apt_tc_data_modeling.apt.apt6_sections import AptFileSectionMetadata

from ifes_apt_tc_data_modeling.apt.apt6_sections_branches import EXPECTED_SECTIONS

from ifes_apt_tc_data_modeling.nexus.nx_field import NxField

from ifes_apt_tc_data_modeling.utils.mmapped_io import get_memory_mapped_data


class ReadAptFileFormat():
    """Read AMETEK's open exchange *.apt file format."""

    def __init__(self, filename: str):
        assert len(filename) > 4, 'APT file incorrect filename ending!'
        assert filename.lower().endswith('.apt'), \
            'APT file incorrect file type!'
        self.filename = filename

        self.filesize = os.path.getsize(self.filename)
        print('Reading ' + self.filename + ' which is ' + str(self.filesize) + ' bytes')

        self.header_section = None
        self.byte_offsets: dict = {}
        self.available_sections: dict = {}
        # where do sections start bytes from beginning of the file?
        self.apt: dict = {}

        self.parse_file_structure()

    def parse_file_structure(self):
        """Parse APT file header plus flat collection of metadata/data pairs.

        Each pair has a so-called section header and a corresponding raw data
        block. Section headers detail the content of the immediately trailing
        raw data block.
        An APT file can store none, some, or all of the possible sections.
        Furthermore the file can contain additional pieces of information
        which this parser cannot read-out because the APT format is maintained
        by AMETEK. The AMETEK source code is the only reliable source of
        information about which content the sections encode and how these
        get formatted when exporting an APT file from APSuite for a specific
        version and build number and type of experiment plus
        combinations of settings.
        Parse header of the file and check which parsable sections the file
        contains get the byte offsets of the sections from the beginning
        relative to the start/beginning of the file.
        """
        self.byte_offsets = {}
        self.header_section = None
        self.available_sections = {}

        with open(self.filename, 'rb') as file_handle:
            self.dummy_header = AptFileHeaderMetadata()
            found_header = np.fromfile(file_handle,
                                       self.dummy_header.get_numpy_struct(),
                                       count=1)

            assert self.dummy_header.matches(found_header), \
                'Found an unexpectedly formatted/versioned header! \
                Please contact the development team to help us inspect \
                the matter.'
            print('File describes ' + str(found_header['llIonCount'][0]) + ' ions')

            self.header_section = found_header
            self.byte_offsets['header'] = np.uint64(file_handle.tell())
            print(self.byte_offsets['header'])

            end_of_file_not_reached = b'yes'
            while end_of_file_not_reached != b'':
                # probe for end of file
                end_of_file_not_reached = file_handle.read(1)
                if end_of_file_not_reached != b'':
                    file_handle.seek(-1, os.SEEK_CUR)
                else:
                    print('End of file at ' + str(file_handle.tell()) + ' bytes')
                    break

                dummy_section = AptFileSectionMetadata()
                found_section = np.fromfile(file_handle,
                                            dummy_section.get_numpy_struct(),
                                            count=1)
                keyword = np_uint16_to_string(
                    found_section['wcSectionType'][0])

                print(keyword)
                print(found_section)
                assert keyword not in self.available_sections.keys(), \
                    'Found a duplicate of an already parsed section! Please \
                    contact the development team as we have never encountered \
                    an example of such a section duplication and here seems \
                    to be an example to inspect the matter.'

                if keyword not in ['Delta Pulse', 'Epos ToF']:
                    assert keyword in EXPECTED_SECTIONS.keys(), \
                        'Found an unknown section, seems like an unknown/new \
                        branch! Please contact the development team to enable us \
                        to contact AMETEK and discuss the situation.'

                    metadata_section = EXPECTED_SECTIONS[keyword]
                    if metadata_section.matches(found_section) is True:
                        # assert metadata_section.matches(found_section), \
                        #     'Found an uninterpretable section! Please contact the \
                        #     development team to help us fixing this.'
                        self.available_sections[keyword] = metadata_section
                else:
                    print('WARNING:: Found an uninterpretable section!')
                    print('WARNING:: This section was not be registered!')
                    print('WARNING:: Please contact the development team!')
                    print('WARNING::     to help us improving this!')
                    print('WARNING:: Try to continue parsing though...!')
                    print('llByteCount ' + str(found_section['llByteCount'][0]))

                self.byte_offsets[keyword] = np.uint64(file_handle.tell())
                if keyword == 'Position':
                    # special case six IEEE 32-bit floats preceeding raw data
                    self.byte_offsets[keyword] += np.uint64(6 * 4)
                self.byte_offsets[keyword] += np.uint64(
                    found_section['llByteCount'][0])
                print('Byte offset for reading data for section: ' + keyword)
                print(self.byte_offsets[keyword])
                # print(file_handle.tell())
                file_handle.seek(self.byte_offsets[keyword], os.SEEK_SET)
                # print(file_handle.tell())

    # one convenience reader function for every known section
    # is useful because it structures the parsers, enables reading the file
    # partially and reduces main memory consumption during full parsing
    def get_header(self):
        """Report metadata in the header."""
        metadata_dict = {
            'cSignature':
                np_uint16_to_string(self.header_section['cSignature'][0]),
            'iHeaderSize':
                np.int32(self.header_section['iHeaderSize'][0]),
            'iHeaderVersion':
                np.int32(self.header_section['iHeaderVersion'][0]),
            'wcFilename':
                np_uint16_to_string(self.header_section['wcFilename'][0]),
            'ftCreationTime':
                np.uint64(self.header_section['ftCreationTime'][0]),
            'llIonCount':
                np.uint64(self.header_section['llIonCount'][0])}
        # check e.g. https://gist.github.com/Mostafa-Hamdy-Elgiar/
        # 9714475f1b3bc224ea063af81566d873 repo
        # for converting Windows/MSDN time to Python time
        for key, value in iter(metadata_dict.items()):
            print(key + ': ' + str(value))

    def get_metadata(self, keyword: str):
        """Report available metadata for quantity if it exists."""
        if keyword in self.available_sections.keys() \
           and keyword in self.byte_offsets.keys():
            metadata_dict = self.available_sections[keyword].get_metadata()
            for key, value in iter(metadata_dict.items()):
                print(key + ': ' + str(value))

    def get_metadata_table(self):
        """Create table from all metadata for each section."""
        column_names = ['section']  # header
        assert 'Mass' in self.available_sections.keys(), \
            'Cannot create table, Mass section not available to guide \
                the creation of the table header!'
        for key in self.available_sections['Mass'].get_metadata().keys():
            column_names.append(key)
        data_frame = pd.DataFrame(columns=column_names)

        for keyword, value in self.available_sections.items():
            row_dct = {'section': keyword}
            # print(f"{keyword}")
            row_dct = {**row_dct, **value.get_metadata()}
            # print(value.get_metadata())
            row_df = pd.DataFrame(row_dct, index=[0])
            # print(row_df)
            data_frame = pd.concat([data_frame, row_df], ignore_index=True)

        data_frame.style.format(precision=3, thousands=",", decimal=".") \
            .format_index(str.upper, axis=1)
        return data_frame

    def get_named_quantity(self, keyword: str):
        """Read quantity with name in keyword from APT file if it exists."""
        if keyword in self.available_sections.keys() \
           and keyword in self.byte_offsets.keys():
            byte_position_start = self.byte_offsets[keyword] \
                - self.available_sections[keyword].get_ametek_size()
            print('Reading section ' + keyword + ' at ' + str(byte_position_start))

            dtype = self.available_sections[keyword].get_ametek_type()
            offset = byte_position_start
            stride = np.uint64(
                self.available_sections[keyword].meta['i_data_type_size'] / 8)
            count = self.available_sections[keyword].get_ametek_count()

            data = get_memory_mapped_data(
                self.filename, dtype, offset, stride, count)

            shape = tuple(self.available_sections[keyword].get_ametek_shape())
            unit = self.available_sections[keyword].meta['wc_data_unit']

            return NxField(
                np.reshape(data, newshape=shape), np_uint16_to_string(unit))

        return NxField()

    def get_mass_to_charge_state_ratio(self):
        """Read mass-to-charge."""
        return self.get_named_quantity('Mass')

    def get_reconstructed_positions(self):
        """Read reconstructed positions."""
        return self.get_named_quantity('Position')

# test cases how to use the parser
# TEST_FILE_NAME = '70_50_50.apt'  # Xuyang Zhou's (MPIE) \
# apt = ReadAptFileFormat(TEST_FILE_NAME)
# print(apt.get_metadata_table())
# print(apt.get_header())
# xyz = apt.get_reconstructed_positions()
# equivalent to
# xyz = apt.get_metadata('Position')
# mq = apt.get_mass_to_charge_state_ratios()
# equivalent to
# mq = parsedFile.get_named_quantity('Mass')



# Utilities for parsing data and molecular ions in atom probe microscopy.
#
# Also convenience functions are included which translate human-readable ion
# names into the isotope_vector description proposed by Kuehbach et al. in
# DOI: 10.1017/S1431927621012241 to the human-readable ion names which are use
# in P. Felfer et al.'s atom probe toolbox
#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

# pylint: disable=no-member,duplicate-code

import re

import typing

from typing import Tuple

import mmap

import numpy as np

from ase.data import atomic_numbers, chemical_symbols

from ifes_apt_tc_data_modeling.utils.nist_isotope_data import isotopes

from ifes_apt_tc_data_modeling.utils.definitions import MAX_NUMBER_OF_ION_SPECIES
from ifes_apt_tc_data_modeling.utils.definitions import MAX_NUMBER_OF_ATOMS_PER_ION
from ifes_apt_tc_data_modeling.utils.definitions import MQ_EPSILON


def isotope_to_hash(proton_number: int = 0,
                 neutron_number: int = 0) -> int:
    """Encode an isotope to a hashvalue."""
    n_protons = np.uint16(proton_number)
    n_neutrons = np.uint16(neutron_number)
    assert (n_protons >= np.uint16(0)) and (n_protons < np.uint16(256)), \
        "Argument proton number on [0, 256) needed!"
    assert (n_neutrons >= np.uint16(0)) and (n_neutrons < np.uint16(256)), \
        "Argument neutron number on [0, 256) needed!"
    hashvalue = int(n_protons + (np.uint16(256) * n_neutrons))
    return hashvalue


def hash_to_isotope(hashvalue: int = 0) -> Tuple[int, int]:
    """Decode a hashvalue to an isotope."""
    # assert isinstance(hashvalue, int), \
    #     "Argument hashvalue needs to be integer!"
    val = np.uint16(hashvalue)
    assert val >= np.uint16(0), \
        "Argument hashvalue needs to be an unsigned integer!"
    assert val <= np.iinfo(np.uint16).max, \
        "Argument hashvalue needs to map on an uint16!"
    neutron_number = np.uint16(val / np.uint16(256))
    proton_number = np.uint16(val - neutron_number * np.uint16(256))
    return (int(proton_number), int(neutron_number))


def create_isotope_vector(building_blocks: list) -> np.ndarray:
    """Create specifically-shaped array of isotope hashvalues."""
    # building_blocks are usually names of elements in the periodic tables
    # if not we assume the ion is special, a user type

    # test cases:
    # create_isotope_vector(["Fe", "Fe", "O", "O", "O"])
    symbol_to_proton_number = atomic_numbers

    ivec = np.zeros((MAX_NUMBER_OF_ATOMS_PER_ION,), np.uint16)
    hashvector = []
    assert len(building_blocks) <= MAX_NUMBER_OF_ATOMS_PER_ION, \
        "Faced an ion with an unsupported high complexity!"
    # MAX_NUMBER_OF_ATOMS_PER_ION can be modified to describe large fragments

    if len(building_blocks) == 0:  # special case unknown ion type
        return ivec

    for block in building_blocks:
        assert isinstance(block, str), \
            "Argument block has to be a string, symbol for an element like K or isotope K-40!"
        assert block != "", \
            "Argument block has to be a non-empty string!"
        if block.count("-") == 0:
            assert (block in symbol_to_proton_number) and (block != "X"), \
                block + " is not a valid chemical symbol!"
            proton_number = symbol_to_proton_number[block]
            neutron_number = 0
            hashvector.append(isotope_to_hash(proton_number, neutron_number))
        elif block.count("-") == 1:
            symb_mass = block.split("-")
            assert len(symb_mass) == 2, \
                block + " is not properly formatted <symbol>-<mass_number>!"
            assert (symb_mass[0] in symbol_to_proton_number) and (symb_mass[0] != "X"), \
                symb_mass[0] + " is not a valid chemical symbol!"
            proton_number = symbol_to_proton_number[symb_mass[0]]
            mass_number = int(symb_mass[1])
            neutron_number = mass_number - proton_number
            assert proton_number in isotopes.keys(), \
                "No isotopes for proton_number " + str(proton_number) + " via ase!"
            assert mass_number in isotopes[proton_number].keys(), \
                "No isotope for mass_number " + str(mass_number) + " via ase!"
            hashvector.append(isotope_to_hash(proton_number, neutron_number))
        else:
            print(f"WARNING: {block} does not specify a unique element name!")
            return ivec

    ivec[0:len(hashvector)] = np.sort(
        np.asarray(hashvector, np.uint16), kind="stable")[::-1]
    return ivec


def isotope_vector_to_nuclid_list(ivec: np.ndarray) -> np.ndarray:
    """Create a NeXus NXion nuclid list."""
    assert np.shape(ivec) == (MAX_NUMBER_OF_ATOMS_PER_ION,), \
        "Argument isotope_vector needs to be shaped (MAX_NUMBER_OF_ATOMS_PER_ION,) !"
    nuclid_list = np.zeros((2, MAX_NUMBER_OF_ATOMS_PER_ION), np.uint16)
    for idx in np.arange(0, MAX_NUMBER_OF_ATOMS_PER_ION):
        if ivec[idx] != 0:
            protons, neutrons = hash_to_isotope(int(ivec[idx]))
            nuclid_list[0, idx] = protons + neutrons
            nuclid_list[1, idx] = protons
    return nuclid_list


def isotope_vector_to_dict_keyword(ivec: np.ndarray) -> str:
    """Create keyword for dictionary from isotope_vector."""
    assert len(ivec) <= MAX_NUMBER_OF_ATOMS_PER_ION, \
        "Argument isotope_vector len <= MAX_NUMBER_OF_ATOMS_PER_ION !"
    lst = []
    for hashvalue in ivec:
        if hashvalue != 0:
            lst.append(str(hashvalue))
    if lst != []:
        return "_".join(lst)
    return "0"  # "_".join(np.asarray(np.zeros((MAX_NUMBER_OF_ATOMS_PER_ION,)), np.uint16))


def isotope_vector_to_human_readable_name(ivec: np.ndarray, charge_state: np.int8) -> str:
    """Get human-readable name from an isotope_vector."""
    assert len(ivec) <= MAX_NUMBER_OF_ATOMS_PER_ION, \
        "Argument isotope_vector len <= MAX_NUMBER_OF_ATOMS_PER_ION !"
    human_readable = ""
    if np.sum(ivec) != 0:
        for hashvalue in ivec:
            if hashvalue != 0:
                protons, neutrons = hash_to_isotope(int(hashvalue))
                if neutrons > 0:
                    human_readable += str(protons + neutrons) \
                        + chemical_symbols[protons]
                else:
                    human_readable += chemical_symbols[protons]
                human_readable += " "
        if 0 < charge_state < 8:
            human_readable += "+" * charge_state
        elif -8 < charge_state < 0:
            human_readable += "-" * -charge_state
        else:
            human_readable = human_readable.rstrip()
        return human_readable
    else:
        return "unknown_iontype"


def is_range_overlapping(interval: np.ndarray,
                        interval_set: np.float64) -> bool:
    """Check if interval overlaps within with members of interval set."""
    assert np.shape(interval) == (2,), "Interval needs to have two columns!"
    assert np.shape(interval_set)[1] == 2, \
        "Interval_set needs to have two columns!"
    # interval = np.array([53.789, 54.343])
    # interval_set = np.array([[27.778, 28.33]])  # for testing purposes
    if np.shape(interval_set)[0] >= 1:
        left_and_right_delta = np.zeros([np.shape(interval_set)[0], 2], bool)
        left_and_right_delta[:, 0] = (interval_set[:, 0] - interval[1]) \
            > MQ_EPSILON
        left_and_right_delta[:, 1] = (interval[0] - interval_set[:, 1]) \
            > MQ_EPSILON
        no_overlap = np.array([np.any(i) for i in left_and_right_delta])
        if np.all(no_overlap):
            return False
        return True
    return False


def is_range_significant(left: np.float64, right: np.float64) -> bool:
    """Check if inclusive interval bounds [left, right] span a finite range."""
    assert left >= np.float64(0.) and right >= np.float64(0.), \
        "Left and right bound have to be positive!"
    if (right - left) > MQ_EPSILON:
        return True
    return False



# RNG range file reader used by atom probe microscopists.
#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

# pylint: disable=no-member,duplicate-code

import re

import numpy as np

from ifes_apt_tc_data_modeling.nexus.nx_ion import NxField, NxIon
from ifes_apt_tc_data_modeling.utils.utils import \
    create_isotope_vector, is_range_significant
from ifes_apt_tc_data_modeling.utils.definitions import MQ_EPSILON
from ifes_apt_tc_data_modeling.utils.molecular_ions import MolecularIonBuilder
from ifes_apt_tc_data_modeling.utils.molecular_ions import \
    PRACTICAL_ABUNDANCE, PRACTICAL_ABUNDANCE_PRODUCT, \
    PRACTICAL_MIN_HALF_LIFE, VERBOSE, SACRIFICE_ISOTOPIC_UNIQUENESS

from ase.data import atomic_numbers, atomic_masses, chemical_symbols


def evaluate_rng_range_line(
        i: int, line: str, column_id_to_label: dict, n_columns: int) -> dict:
    """Represent information content of a single range line."""
    # example line: ". 107.7240 108.0960 1 0 0 0 0 0 0 0 0 0 3 0 0 0"
    info: dict = {}
    info["identifier"] = "Range" + str(i)
    info["range"] = np.asarray([0., MQ_EPSILON], np.float64)
    info["atoms"] = []
    info["volume"] = np.float64(0.)
    info["color"] = ""
    info["name"] = ""

    tmp = re.split(r"\s+", line)
    assert len(tmp) is n_columns, "Line " + line \
        + " inconsistent number columns!"
    assert tmp[0] == ".", "Line " + line \
        + " has inconsistent line prefix!"

    assert is_range_significant(np.float64(tmp[1]), np.float64(tmp[2])), \
        "Line " + line + " insignificant range!"
    info["range"] = np.asarray([tmp[1], tmp[2]], np.float64)

    # line encodes multiplicity of element via array of multiplicity counts
    element_multiplicity = np.asarray(tmp[3:len(tmp)], np.uint32)
    assert np.sum(element_multiplicity) > 0, \
        "Line " + line + " no element counts!"
    if np.sum(element_multiplicity) > 0:
        for j in np.arange(0, len(element_multiplicity)):
            assert element_multiplicity[j] >= 0, "Line " + line \
               + " no negative element counts!"
            if element_multiplicity[j] > 0:
                symbol = column_id_to_label[j + 1]
                if (symbol in chemical_symbols) and (symbol != "X"):
                    info["atoms"] = np.append(
                        info["atoms"], [column_id_to_label[j + 1]] \
                        * int(element_multiplicity[j]))
                else:
                    info["name"] = symbol
                    info["atoms"] = []  # will map to unknown type

    # color for RNG files can only be decoded by
    # loading the color of elements and polyatomic extensions
    # and then check to which category (element or polyatomic) a range
    # belongs and then take this color, there is almost no way
    # to make something so simple for disentangled
    return info


def evaluate_rng_ion_type_header(line: str) -> dict:
    """Represent information content in the key header line."""
    # line = "------------------- Fe Mg Al Mn Si V C Ga Ti Ca O Na Co H"
    # line = "---- a"
    # line = "----------------- Sc Fe O C Al Si Cr H unknown"
    info = {}
    info["column_id_to_label"] = {}
    tmp = re.split(r"\s+", line)
    assert len(tmp) > 1, "RNG file does not contain iontype labels!"
    for i in np.arange(1, len(tmp)):
        info["column_id_to_label"][i] = tmp[i]
    return info


class ReadRngFileFormat():
    """Read *.rng file format."""

    def __init__(self, filename: str):
        assert len(filename) > 4, "RNG file incorrect filename ending!"
        assert filename.lower().endswith(".rng"), \
            "RNG file incorrect file type!"
        self.filename = filename

        self.rng: dict = {}
        self.rng["ranges"] = {}
        self.rng["ions"] = {}
        self.rng["molecular_ions"] = []
        self.read_rng()

    def read_rng(self):
        """Read RNG range file content."""
        with open(self.filename, mode="r", encoding="utf8") as rngf:
            txt = rngf.read()

        txt = txt.replace("\r\n", "\n")  # windows to unix EOL conversion
        txt = txt.replace(",", ".")  # use decimal dots instead of comma
        txt_stripped = [line for line in txt.split("\n")
                        if line.strip() != "" and line.startswith("#") is False]
        del txt

        # see DOI: 10.1007/978-1-4899-7430-3 for further details to this
        # Oak Ridge National Lab / Oxford *.rng file format
        # only the first ------ line is relevant
        # it details all ion labels aka ions
        # AMETEK"s IVAS/APSuite-specific trailing
        # polyatomic extension is redundant info

        tmp = None
        current_line_id = int(0)  # search key header line
        for line in txt_stripped:
            tmp = re.search(r"----", line)
            if tmp is None:
                current_line_id += int(1)
            else:
                break
        assert tmp is not None, "RNG file does not contain key header line!"

        header = evaluate_rng_ion_type_header(txt_stripped[current_line_id])

        tmp = re.split(r"\s+", txt_stripped[0])
        assert tmp[0].isnumeric() is True, "Number of species corrupted!"
        n_element_symbols = int(tmp[0])
        assert n_element_symbols >= 0, "No species defined!"
        assert tmp[1].isnumeric() is True, "Number of ranges corrupted!"
        n_ranges = int(tmp[1])
        assert n_ranges >= 0, "No ranges defined!"

        for i in np.arange(current_line_id + 1,
                           current_line_id + 1 + n_ranges):
            dct = evaluate_rng_range_line(
                i - current_line_id, txt_stripped[i],
                header["column_id_to_label"],
                n_element_symbols + 3)
            assert dct is not None, \
                "Line " + txt_stripped[i] + " is corrupted!"

            m_ion = NxIon(isotope_vector=create_isotope_vector(
                dct["atoms"]), charge_state=0)
            m_ion.add_range(dct["range"][0], dct["range"][1])
            m_ion.comment = NxField(dct["name"], "")
            m_ion.color = NxField(dct["color"], "")
            m_ion.volume = NxField(dct["volume"], "")
            # m_ion.report()

            crawler = MolecularIonBuilder(
                min_abundance=PRACTICAL_ABUNDANCE,
                min_abundance_product=PRACTICAL_ABUNDANCE_PRODUCT,
                min_half_life=PRACTICAL_MIN_HALF_LIFE,
                sacrifice_uniqueness=SACRIFICE_ISOTOPIC_UNIQUENESS,
                verbose=VERBOSE)
            recovered_charge_state, m_ion_candidates = crawler.combinatorics(
                m_ion.isotope_vector.typed_value,
                m_ion.ranges.typed_value[0, 0],
                m_ion.ranges.typed_value[0, 1])
            # print(f"{recovered_charge_state}")
            m_ion.charge_state = NxField(np.int8(recovered_charge_state), "")
            m_ion.update_human_readable_name()
            m_ion.add_charge_state_model(
                {"min_abundance": PRACTICAL_ABUNDANCE,
                 "min_abundance_product": PRACTICAL_ABUNDANCE_PRODUCT,
                 "min_half_life": PRACTICAL_MIN_HALF_LIFE,
                 "sacrifice_isotopic_uniqueness": SACRIFICE_ISOTOPIC_UNIQUENESS},
                m_ion_candidates)

            self.rng["molecular_ions"].append(m_ion)
        print(self.filename + " parsed successfully")

if __name__ == "main":
    pass
    # testing
    # parsedFile = ReadRngFileFormat("../../SeHoKim_R5076_44076_v02.rng")



# AMETEK APT(6) data exchange file reader used by atom probe microscopists.
#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

# pylint: disable=no-member,duplicate-code

import numpy as np


def np_uint16_to_string(uint16_array: np.ndarray) -> str:
    """Create string from array of uint16 numbers (UTF-16)."""
    str_parsed = ''
    for value in uint16_array:
        if value != 0:  # '\x00'
            str_parsed += chr(value)
    return str(str_parsed)


def string_to_typed_nparray(string: str, length: int,
                            dtype: type) -> np.ndarray:
    """Create fixed length np.uint16 numpy array from string."""
    assert dtype is not None, 'dtype must not be None!'
    assert len(string) <= length, 'Input string is longer than \
                                  number of array elements !'
    nparray = np.zeros(length, dtype)
    for value in np.arange(0, len(string)):
        nparray[value] = ord(string[value])
    return nparray



# AMETEK APT(6) data exchange file reader used by atom probe microscopists.
#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

# pylint: disable=no-member,duplicate-code

from ifes_apt_tc_data_modeling.apt.apt6_sections import AptFileSectionMetadata


# define which header and sections to expect in an *.apt file
# the sections below are referred to as branches in the commercial software
# APSuite enables users select prior I/O which of the sections to write
# normally a range file from based on the exchange with AMETEK and

# in addition there is usually a sextet of preceeding IEEE 32-bit floats
# to the position section. This sextett encodes the min,max bounds of the point
# cloud in x, y, z direction respectively
# F. M. M. de Oliveira reported cases where *.apt from mere flat test runs,
# i.e. software sessions during which no reconstruction is performed
# this sextett is absent

EXPECTED_SECTIONS = {}

# most tested sections

EXPECTED_SECTIONS['Position'] = AptFileSectionMetadata()
EXPECTED_SECTIONS['Position'].set_section_name('Position')
EXPECTED_SECTIONS['Position'].set_c_signature()
EXPECTED_SECTIONS['Position'].set_i_header_size(148 + 6 * 4)
# six 4 byte IEEE 32-bit floats are following immediately after a 'Position'
# section trailing the actual position value array
EXPECTED_SECTIONS['Position'].set_i_header_version(2)
EXPECTED_SECTIONS['Position'].set_wc_section_type('Position')
EXPECTED_SECTIONS['Position'].set_i_section_version(1)
EXPECTED_SECTIONS['Position'].set_e_relationship_type(1)
EXPECTED_SECTIONS['Position'].set_e_record_type(1)
EXPECTED_SECTIONS['Position'].set_e_record_data_type(3)
EXPECTED_SECTIONS['Position'].set_i_data_type_size(32)
EXPECTED_SECTIONS['Position'].set_i_record_size(12)
EXPECTED_SECTIONS['Position'].set_wc_data_unit('nm')
EXPECTED_SECTIONS['Position'].set_accepted_units(['nm'])

EXPECTED_SECTIONS['Mass'] = AptFileSectionMetadata()
EXPECTED_SECTIONS['Mass'].set_section_name('Mass')
EXPECTED_SECTIONS['Mass'].set_c_signature()
EXPECTED_SECTIONS['Mass'].set_i_header_size(148)
EXPECTED_SECTIONS['Mass'].set_i_header_version(2)
EXPECTED_SECTIONS['Mass'].set_wc_section_type('Mass')
EXPECTED_SECTIONS['Mass'].set_i_section_version(1)
EXPECTED_SECTIONS['Mass'].set_e_relationship_type(1)
EXPECTED_SECTIONS['Mass'].set_e_record_type(1)
EXPECTED_SECTIONS['Mass'].set_e_record_data_type(3)
EXPECTED_SECTIONS['Mass'].set_i_data_type_size(32)
EXPECTED_SECTIONS['Mass'].set_i_record_size(4)
EXPECTED_SECTIONS['Mass'].set_wc_data_unit('Da')
EXPECTED_SECTIONS['Mass'].set_accepted_units(['da', 'Da', 'amu'])

# M. K\"uhbach detected there are at least APSuite versions which write
# files with Da and some with da as wcDataUnits argument, it seems there
# was a source code change, we use pint to detect these issues and warn

# sections which demand more testing with the parser

EXPECTED_SECTIONS['z'] = AptFileSectionMetadata()
EXPECTED_SECTIONS['z'].set_section_name('z')
EXPECTED_SECTIONS['z'].set_c_signature()
EXPECTED_SECTIONS['z'].set_i_header_size(148)
EXPECTED_SECTIONS['z'].set_i_header_version(2)
EXPECTED_SECTIONS['z'].set_wc_section_type('z')
EXPECTED_SECTIONS['z'].set_i_section_version(1)
EXPECTED_SECTIONS['z'].set_e_relationship_type(1)
EXPECTED_SECTIONS['z'].set_e_record_type(1)
EXPECTED_SECTIONS['z'].set_e_record_data_type(1)
EXPECTED_SECTIONS['z'].set_i_data_type_size(64)
EXPECTED_SECTIONS['z'].set_i_record_size(8)
EXPECTED_SECTIONS['z'].set_wc_data_unit('ions')
EXPECTED_SECTIONS['z'].set_accepted_units(['ions'])

EXPECTED_SECTIONS['tof'] = AptFileSectionMetadata()
EXPECTED_SECTIONS['tof'].set_section_name('tof')
EXPECTED_SECTIONS['tof'].set_c_signature()
EXPECTED_SECTIONS['tof'].set_i_header_size(148)
EXPECTED_SECTIONS['tof'].set_i_header_version(2)
EXPECTED_SECTIONS['tof'].set_wc_section_type('tof')
EXPECTED_SECTIONS['tof'].set_i_section_version(1)
EXPECTED_SECTIONS['tof'].set_e_relationship_type(1)
EXPECTED_SECTIONS['tof'].set_e_record_type(1)
EXPECTED_SECTIONS['tof'].set_e_record_data_type(3)
EXPECTED_SECTIONS['tof'].set_i_data_type_size(32)
EXPECTED_SECTIONS['tof'].set_i_record_size(4)
EXPECTED_SECTIONS['tof'].set_wc_data_unit('ns')
EXPECTED_SECTIONS['tof'].set_accepted_units(['ns'])

EXPECTED_SECTIONS['Voltage'] = AptFileSectionMetadata()
EXPECTED_SECTIONS['Voltage'].set_section_name('Voltage')
EXPECTED_SECTIONS['Voltage'].set_c_signature()
EXPECTED_SECTIONS['Voltage'].set_i_header_size(148)
EXPECTED_SECTIONS['Voltage'].set_i_header_version(2)
EXPECTED_SECTIONS['Voltage'].set_wc_section_type('Voltage')
EXPECTED_SECTIONS['Voltage'].set_i_section_version(1)
EXPECTED_SECTIONS['Voltage'].set_e_relationship_type(1)
EXPECTED_SECTIONS['Voltage'].set_e_record_type(1)
EXPECTED_SECTIONS['Voltage'].set_e_record_data_type(3)
EXPECTED_SECTIONS['Voltage'].set_i_data_type_size(32)
EXPECTED_SECTIONS['Voltage'].set_i_record_size(4)
EXPECTED_SECTIONS['Voltage'].set_wc_data_unit('')

EXPECTED_SECTIONS['pulse'] = AptFileSectionMetadata()
EXPECTED_SECTIONS['pulse'].set_section_name('pulse')
EXPECTED_SECTIONS['pulse'].set_c_signature()
EXPECTED_SECTIONS['pulse'].set_i_header_size(148)
EXPECTED_SECTIONS['pulse'].set_i_header_version(2)
EXPECTED_SECTIONS['pulse'].set_wc_section_type('pulse')
EXPECTED_SECTIONS['pulse'].set_i_section_version(1)
EXPECTED_SECTIONS['pulse'].set_e_relationship_type(1)
EXPECTED_SECTIONS['pulse'].set_e_record_type(1)
EXPECTED_SECTIONS['pulse'].set_e_record_data_type(3)
EXPECTED_SECTIONS['pulse'].set_i_data_type_size(32)
EXPECTED_SECTIONS['pulse'].set_i_record_size(4)
EXPECTED_SECTIONS['pulse'].set_wc_data_unit('')

EXPECTED_SECTIONS['freq'] = AptFileSectionMetadata()
EXPECTED_SECTIONS['freq'].set_section_name('freq')
EXPECTED_SECTIONS['freq'].set_c_signature()
EXPECTED_SECTIONS['freq'].set_i_header_size(148)
EXPECTED_SECTIONS['freq'].set_i_header_version(2)
EXPECTED_SECTIONS['freq'].set_wc_section_type('freq')
EXPECTED_SECTIONS['freq'].set_i_section_version(1)
EXPECTED_SECTIONS['freq'].set_e_relationship_type(1)
EXPECTED_SECTIONS['freq'].set_e_record_type(1)
EXPECTED_SECTIONS['freq'].set_e_record_data_type(3)
EXPECTED_SECTIONS['freq'].set_i_data_type_size(32)
EXPECTED_SECTIONS['freq'].set_i_record_size(4)
EXPECTED_SECTIONS['freq'].set_wc_data_unit('Hz')
EXPECTED_SECTIONS['freq'].set_accepted_units(['Hz'])

EXPECTED_SECTIONS['tElapsed'] = AptFileSectionMetadata()
EXPECTED_SECTIONS['tElapsed'].set_section_name('tElapsed')
EXPECTED_SECTIONS['tElapsed'].set_c_signature()
EXPECTED_SECTIONS['tElapsed'].set_i_header_size(148)
EXPECTED_SECTIONS['tElapsed'].set_i_header_version(2)
EXPECTED_SECTIONS['tElapsed'].set_wc_section_type('tElapsed')
EXPECTED_SECTIONS['tElapsed'].set_i_section_version(1)
EXPECTED_SECTIONS['tElapsed'].set_e_relationship_type(1)
EXPECTED_SECTIONS['tElapsed'].set_e_record_type(1)
EXPECTED_SECTIONS['tElapsed'].set_e_record_data_type(3)
EXPECTED_SECTIONS['tElapsed'].set_i_data_type_size(32)
EXPECTED_SECTIONS['tElapsed'].set_i_record_size(4)
EXPECTED_SECTIONS['tElapsed'].set_wc_data_unit('')

EXPECTED_SECTIONS['erate'] = AptFileSectionMetadata()
EXPECTED_SECTIONS['erate'].set_section_name('erate')
EXPECTED_SECTIONS['erate'].set_c_signature()
EXPECTED_SECTIONS['erate'].set_i_header_size(148)
EXPECTED_SECTIONS['erate'].set_i_header_version(2)
EXPECTED_SECTIONS['erate'].set_wc_section_type('erate')
EXPECTED_SECTIONS['erate'].set_i_section_version(1)
EXPECTED_SECTIONS['erate'].set_e_relationship_type(1)
EXPECTED_SECTIONS['erate'].set_e_record_type(1)
EXPECTED_SECTIONS['erate'].set_e_record_data_type(3)
EXPECTED_SECTIONS['erate'].set_i_data_type_size(32)
EXPECTED_SECTIONS['erate'].set_i_record_size(4)
EXPECTED_SECTIONS['erate'].set_wc_data_unit('%/100')
EXPECTED_SECTIONS['erate'].set_accepted_units(['%/100'])

EXPECTED_SECTIONS['xstage'] = AptFileSectionMetadata()
EXPECTED_SECTIONS['xstage'].set_section_name('xstage')
EXPECTED_SECTIONS['xstage'].set_c_signature()
EXPECTED_SECTIONS['xstage'].set_i_header_size(148)
EXPECTED_SECTIONS['xstage'].set_i_header_version(2)
EXPECTED_SECTIONS['xstage'].set_wc_section_type('xstage')
EXPECTED_SECTIONS['xstage'].set_i_section_version(1)
EXPECTED_SECTIONS['xstage'].set_e_relationship_type(1)
EXPECTED_SECTIONS['xstage'].set_e_record_type(1)
EXPECTED_SECTIONS['xstage'].set_e_record_data_type(1)
EXPECTED_SECTIONS['xstage'].set_i_data_type_size(32)
EXPECTED_SECTIONS['xstage'].set_i_record_size(4)
EXPECTED_SECTIONS['xstage'].set_wc_data_unit('')

EXPECTED_SECTIONS['ystage'] = AptFileSectionMetadata()
EXPECTED_SECTIONS['ystage'].set_section_name('ystage')
EXPECTED_SECTIONS['ystage'].set_c_signature()
EXPECTED_SECTIONS['ystage'].set_i_header_size(148)
EXPECTED_SECTIONS['ystage'].set_i_header_version(2)
EXPECTED_SECTIONS['ystage'].set_wc_section_type('ystage')
EXPECTED_SECTIONS['ystage'].set_i_section_version(1)
EXPECTED_SECTIONS['ystage'].set_e_relationship_type(1)
EXPECTED_SECTIONS['ystage'].set_e_record_type(1)
EXPECTED_SECTIONS['ystage'].set_e_record_data_type(1)
EXPECTED_SECTIONS['ystage'].set_i_data_type_size(32)
EXPECTED_SECTIONS['ystage'].set_i_record_size(4)
EXPECTED_SECTIONS['ystage'].set_wc_data_unit('')

EXPECTED_SECTIONS['zstage'] = AptFileSectionMetadata()
EXPECTED_SECTIONS['zstage'].set_section_name('zstage')
EXPECTED_SECTIONS['zstage'].set_c_signature()
EXPECTED_SECTIONS['zstage'].set_i_header_size(148)
EXPECTED_SECTIONS['zstage'].set_i_header_version(2)
EXPECTED_SECTIONS['zstage'].set_wc_section_type('zstage')
EXPECTED_SECTIONS['zstage'].set_i_section_version(1)
EXPECTED_SECTIONS['zstage'].set_e_relationship_type(1)
EXPECTED_SECTIONS['zstage'].set_e_record_type(1)
EXPECTED_SECTIONS['zstage'].set_e_record_data_type(1)
EXPECTED_SECTIONS['zstage'].set_i_data_type_size(32)
EXPECTED_SECTIONS['zstage'].set_i_record_size(4)
EXPECTED_SECTIONS['zstage'].set_wc_data_unit('')

EXPECTED_SECTIONS['tstage'] = AptFileSectionMetadata()
EXPECTED_SECTIONS['tstage'].set_section_name('tstage')
EXPECTED_SECTIONS['tstage'].set_c_signature()
EXPECTED_SECTIONS['tstage'].set_i_header_size(148)
EXPECTED_SECTIONS['tstage'].set_i_header_version(2)
EXPECTED_SECTIONS['tstage'].set_wc_section_type('tstage')
EXPECTED_SECTIONS['tstage'].set_i_section_version(1)
EXPECTED_SECTIONS['tstage'].set_e_relationship_type(1)
EXPECTED_SECTIONS['tstage'].set_e_record_type(1)
EXPECTED_SECTIONS['tstage'].set_e_record_data_type(2)
EXPECTED_SECTIONS['tstage'].set_i_data_type_size(16)
EXPECTED_SECTIONS['tstage'].set_i_record_size(2)
EXPECTED_SECTIONS['tstage'].set_wc_data_unit('')

EXPECTED_SECTIONS['TargetErate'] = AptFileSectionMetadata()
EXPECTED_SECTIONS['TargetErate'].set_section_name('TargetErate')
EXPECTED_SECTIONS['TargetErate'].set_c_signature()
EXPECTED_SECTIONS['TargetErate'].set_i_header_size(148)
EXPECTED_SECTIONS['TargetErate'].set_i_header_version(2)
EXPECTED_SECTIONS['TargetErate'].set_wc_section_type('TargetErate')
EXPECTED_SECTIONS['TargetErate'].set_i_section_version(1)
EXPECTED_SECTIONS['TargetErate'].set_e_relationship_type(1)
EXPECTED_SECTIONS['TargetErate'].set_e_record_type(1)
EXPECTED_SECTIONS['TargetErate'].set_e_record_data_type(3)
EXPECTED_SECTIONS['TargetErate'].set_i_data_type_size(32)
EXPECTED_SECTIONS['TargetErate'].set_i_record_size(4)
EXPECTED_SECTIONS['TargetErate'].set_wc_data_unit('%/100')
EXPECTED_SECTIONS['TargetErate'].set_accepted_units(['%/100'])

EXPECTED_SECTIONS['TargetFlux'] = AptFileSectionMetadata()
EXPECTED_SECTIONS['TargetFlux'].set_section_name('TargetFlux')
EXPECTED_SECTIONS['TargetFlux'].set_c_signature()
EXPECTED_SECTIONS['TargetFlux'].set_i_header_size(148)
EXPECTED_SECTIONS['TargetFlux'].set_i_header_version(2)
EXPECTED_SECTIONS['TargetFlux'].set_wc_section_type('TargetFlux')
EXPECTED_SECTIONS['TargetFlux'].set_i_section_version(1)
EXPECTED_SECTIONS['TargetFlux'].set_e_relationship_type(1)
EXPECTED_SECTIONS['TargetFlux'].set_e_record_type(1)
EXPECTED_SECTIONS['TargetFlux'].set_e_record_data_type(3)
EXPECTED_SECTIONS['TargetFlux'].set_i_data_type_size(32)
EXPECTED_SECTIONS['TargetFlux'].set_i_record_size(4)
EXPECTED_SECTIONS['TargetFlux'].set_wc_data_unit('')

EXPECTED_SECTIONS['pulseDelta'] = AptFileSectionMetadata()
EXPECTED_SECTIONS['pulseDelta'].set_section_name('pulseDelta')
EXPECTED_SECTIONS['pulseDelta'].set_c_signature()
EXPECTED_SECTIONS['pulseDelta'].set_i_header_size(148)
EXPECTED_SECTIONS['pulseDelta'].set_i_header_version(2)
EXPECTED_SECTIONS['pulseDelta'].set_wc_section_type('pulseDelta')
EXPECTED_SECTIONS['pulseDelta'].set_i_section_version(1)
EXPECTED_SECTIONS['pulseDelta'].set_e_relationship_type(1)
EXPECTED_SECTIONS['pulseDelta'].set_e_record_type(1)
EXPECTED_SECTIONS['pulseDelta'].set_e_record_data_type(1)
EXPECTED_SECTIONS['pulseDelta'].set_i_data_type_size(16)
EXPECTED_SECTIONS['pulseDelta'].set_i_record_size(2)
EXPECTED_SECTIONS['pulseDelta'].set_wc_data_unit('')

EXPECTED_SECTIONS['Pres'] = AptFileSectionMetadata()
EXPECTED_SECTIONS['Pres'].set_section_name('Pres')
EXPECTED_SECTIONS['Pres'].set_c_signature()
EXPECTED_SECTIONS['Pres'].set_i_header_size(148)
EXPECTED_SECTIONS['Pres'].set_i_header_version(2)
EXPECTED_SECTIONS['Pres'].set_wc_section_type('Pres')
EXPECTED_SECTIONS['Pres'].set_i_section_version(1)
EXPECTED_SECTIONS['Pres'].set_e_relationship_type(1)
EXPECTED_SECTIONS['Pres'].set_e_record_type(1)
EXPECTED_SECTIONS['Pres'].set_e_record_data_type(3)
EXPECTED_SECTIONS['Pres'].set_i_data_type_size(32)
EXPECTED_SECTIONS['Pres'].set_i_record_size(4)
EXPECTED_SECTIONS['Pres'].set_wc_data_unit('torr')
EXPECTED_SECTIONS['Pres'].set_accepted_units(['torr'])

EXPECTED_SECTIONS['VAnodeMon'] = AptFileSectionMetadata()
EXPECTED_SECTIONS['VAnodeMon'].set_section_name('VAnodeMon')
EXPECTED_SECTIONS['VAnodeMon'].set_c_signature()
EXPECTED_SECTIONS['VAnodeMon'].set_i_header_size(148)
EXPECTED_SECTIONS['VAnodeMon'].set_i_header_version(2)
EXPECTED_SECTIONS['VAnodeMon'].set_wc_section_type('VAnodeMon')
EXPECTED_SECTIONS['VAnodeMon'].set_i_section_version(1)
EXPECTED_SECTIONS['VAnodeMon'].set_e_relationship_type(1)
EXPECTED_SECTIONS['VAnodeMon'].set_e_record_type(1)
EXPECTED_SECTIONS['VAnodeMon'].set_e_record_data_type(3)
EXPECTED_SECTIONS['VAnodeMon'].set_i_data_type_size(32)
EXPECTED_SECTIONS['VAnodeMon'].set_i_record_size(4)
EXPECTED_SECTIONS['VAnodeMon'].set_wc_data_unit('')

EXPECTED_SECTIONS['Temp'] = AptFileSectionMetadata()
EXPECTED_SECTIONS['Temp'].set_section_name('Temp')
EXPECTED_SECTIONS['Temp'].set_c_signature()
EXPECTED_SECTIONS['Temp'].set_i_header_size(148)
EXPECTED_SECTIONS['Temp'].set_i_header_version(2)
EXPECTED_SECTIONS['Temp'].set_wc_section_type('Temp')
EXPECTED_SECTIONS['Temp'].set_i_section_version(1)
EXPECTED_SECTIONS['Temp'].set_e_relationship_type(1)
EXPECTED_SECTIONS['Temp'].set_e_record_type(1)
EXPECTED_SECTIONS['Temp'].set_e_record_data_type(3)
EXPECTED_SECTIONS['Temp'].set_i_data_type_size(32)
EXPECTED_SECTIONS['Temp'].set_i_record_size(4)
EXPECTED_SECTIONS['Temp'].set_wc_data_unit('K')
EXPECTED_SECTIONS['Temp'].set_accepted_units(['K'])


EXPECTED_SECTIONS['AmbTemp'] = AptFileSectionMetadata()
EXPECTED_SECTIONS['AmbTemp'].set_section_name('AmbTemp')
EXPECTED_SECTIONS['AmbTemp'].set_c_signature()
EXPECTED_SECTIONS['AmbTemp'].set_i_header_size(148)
EXPECTED_SECTIONS['AmbTemp'].set_i_header_version(2)
EXPECTED_SECTIONS['AmbTemp'].set_wc_section_type('AmbTemp')
EXPECTED_SECTIONS['AmbTemp'].set_i_section_version(1)
EXPECTED_SECTIONS['AmbTemp'].set_e_relationship_type(1)
EXPECTED_SECTIONS['AmbTemp'].set_e_record_type(1)
EXPECTED_SECTIONS['AmbTemp'].set_e_record_data_type(3)
EXPECTED_SECTIONS['AmbTemp'].set_i_data_type_size(32)
EXPECTED_SECTIONS['AmbTemp'].set_i_record_size(4)
EXPECTED_SECTIONS['AmbTemp'].set_wc_data_unit('C')
EXPECTED_SECTIONS['AmbTemp'].set_accepted_units(['C'])

EXPECTED_SECTIONS['laserx'] = AptFileSectionMetadata()
EXPECTED_SECTIONS['laserx'].set_section_name('laserx')
EXPECTED_SECTIONS['laserx'].set_c_signature()
EXPECTED_SECTIONS['laserx'].set_i_header_size(148)
EXPECTED_SECTIONS['laserx'].set_i_header_version(2)
EXPECTED_SECTIONS['laserx'].set_wc_section_type('laserx')
EXPECTED_SECTIONS['laserx'].set_i_section_version(1)
EXPECTED_SECTIONS['laserx'].set_e_relationship_type(1)
EXPECTED_SECTIONS['laserx'].set_e_record_type(1)
EXPECTED_SECTIONS['laserx'].set_e_record_data_type(1)
EXPECTED_SECTIONS['laserx'].set_i_data_type_size(32)
EXPECTED_SECTIONS['laserx'].set_i_record_size(4)
EXPECTED_SECTIONS['laserx'].set_wc_data_unit('')

EXPECTED_SECTIONS['lasery'] = AptFileSectionMetadata()
EXPECTED_SECTIONS['lasery'].set_section_name('lasery')
EXPECTED_SECTIONS['lasery'].set_c_signature()
EXPECTED_SECTIONS['lasery'].set_i_header_size(148)
EXPECTED_SECTIONS['lasery'].set_i_header_version(2)
EXPECTED_SECTIONS['lasery'].set_wc_section_type('lasery')
EXPECTED_SECTIONS['lasery'].set_i_section_version(1)
EXPECTED_SECTIONS['lasery'].set_e_relationship_type(1)
EXPECTED_SECTIONS['lasery'].set_e_record_type(1)
EXPECTED_SECTIONS['lasery'].set_e_record_data_type(1)
EXPECTED_SECTIONS['lasery'].set_i_data_type_size(32)
EXPECTED_SECTIONS['lasery'].set_i_record_size(4)
EXPECTED_SECTIONS['lasery'].set_wc_data_unit('')

EXPECTED_SECTIONS['laserz'] = AptFileSectionMetadata()
EXPECTED_SECTIONS['laserz'].set_section_name('laserz')
EXPECTED_SECTIONS['laserz'].set_c_signature()
EXPECTED_SECTIONS['laserz'].set_i_header_size(148)
EXPECTED_SECTIONS['laserz'].set_i_header_version(2)
EXPECTED_SECTIONS['laserz'].set_wc_section_type('laserz')
EXPECTED_SECTIONS['laserz'].set_i_section_version(1)
EXPECTED_SECTIONS['laserz'].set_e_relationship_type(1)
EXPECTED_SECTIONS['laserz'].set_e_record_type(1)
EXPECTED_SECTIONS['laserz'].set_e_record_data_type(1)
EXPECTED_SECTIONS['laserz'].set_i_data_type_size(32)
EXPECTED_SECTIONS['laserz'].set_i_record_size(4)
EXPECTED_SECTIONS['laserz'].set_wc_data_unit('')

EXPECTED_SECTIONS['laserpower'] = AptFileSectionMetadata()
EXPECTED_SECTIONS['laserpower'].set_section_name('laserpower')
EXPECTED_SECTIONS['laserpower'].set_c_signature()
EXPECTED_SECTIONS['laserpower'].set_i_header_size(148)
EXPECTED_SECTIONS['laserpower'].set_i_header_version(2)
EXPECTED_SECTIONS['laserpower'].set_wc_section_type('laserpower')
EXPECTED_SECTIONS['laserpower'].set_i_section_version(1)
EXPECTED_SECTIONS['laserpower'].set_e_relationship_type(1)
EXPECTED_SECTIONS['laserpower'].set_e_record_type(1)
EXPECTED_SECTIONS['laserpower'].set_e_record_data_type(3)
EXPECTED_SECTIONS['laserpower'].set_i_data_type_size(32)
EXPECTED_SECTIONS['laserpower'].set_i_record_size(4)
EXPECTED_SECTIONS['laserpower'].set_wc_data_unit('')

EXPECTED_SECTIONS['FractureGuard'] = AptFileSectionMetadata()
EXPECTED_SECTIONS['FractureGuard'].set_section_name('FractureGuard')
EXPECTED_SECTIONS['FractureGuard'].set_c_signature()
EXPECTED_SECTIONS['FractureGuard'].set_i_header_size(148)
EXPECTED_SECTIONS['FractureGuard'].set_i_header_version(2)
EXPECTED_SECTIONS['FractureGuard'].set_wc_section_type('FractureGuard')
EXPECTED_SECTIONS['FractureGuard'].set_i_section_version(1)
EXPECTED_SECTIONS['FractureGuard'].set_e_relationship_type(1)
EXPECTED_SECTIONS['FractureGuard'].set_e_record_type(1)
EXPECTED_SECTIONS['FractureGuard'].set_e_record_data_type(2)
EXPECTED_SECTIONS['FractureGuard'].set_i_data_type_size(16)
EXPECTED_SECTIONS['FractureGuard'].set_i_record_size(2)
EXPECTED_SECTIONS['FractureGuard'].set_wc_data_unit('')

EXPECTED_SECTIONS['Noise'] = AptFileSectionMetadata()
EXPECTED_SECTIONS['Noise'].set_section_name('Noise')
EXPECTED_SECTIONS['Noise'].set_c_signature()
EXPECTED_SECTIONS['Noise'].set_i_header_size(148)
EXPECTED_SECTIONS['Noise'].set_i_header_version(2)
EXPECTED_SECTIONS['Noise'].set_wc_section_type('Noise')
EXPECTED_SECTIONS['Noise'].set_i_section_version(1)
EXPECTED_SECTIONS['Noise'].set_e_relationship_type(1)
EXPECTED_SECTIONS['Noise'].set_e_record_type(1)
EXPECTED_SECTIONS['Noise'].set_e_record_data_type(3)
EXPECTED_SECTIONS['Noise'].set_i_data_type_size(32)
EXPECTED_SECTIONS['Noise'].set_i_record_size(4)
EXPECTED_SECTIONS['Noise'].set_wc_data_unit('ions')
EXPECTED_SECTIONS['Noise'].set_accepted_units(['', 'ions'])

EXPECTED_SECTIONS['Uniformity'] = AptFileSectionMetadata()
EXPECTED_SECTIONS['Uniformity'].set_section_name('Uniformity')
EXPECTED_SECTIONS['Uniformity'].set_c_signature()
EXPECTED_SECTIONS['Uniformity'].set_i_header_size(148)
EXPECTED_SECTIONS['Uniformity'].set_i_header_version(2)
EXPECTED_SECTIONS['Uniformity'].set_wc_section_type('Uniformity')
EXPECTED_SECTIONS['Uniformity'].set_i_section_version(1)
EXPECTED_SECTIONS['Uniformity'].set_e_relationship_type(1)
EXPECTED_SECTIONS['Uniformity'].set_e_record_type(1)
EXPECTED_SECTIONS['Uniformity'].set_e_record_data_type(3)
EXPECTED_SECTIONS['Uniformity'].set_i_data_type_size(32)
EXPECTED_SECTIONS['Uniformity'].set_i_record_size(4)
EXPECTED_SECTIONS['Uniformity'].set_wc_data_unit('')

EXPECTED_SECTIONS['tofc'] = AptFileSectionMetadata()
EXPECTED_SECTIONS['tofc'].set_section_name('tofc')
EXPECTED_SECTIONS['tofc'].set_c_signature()
EXPECTED_SECTIONS['tofc'].set_i_header_size(148)
EXPECTED_SECTIONS['tofc'].set_i_header_version(2)
EXPECTED_SECTIONS['tofc'].set_wc_section_type('tofc')
EXPECTED_SECTIONS['tofc'].set_i_section_version(1)
EXPECTED_SECTIONS['tofc'].set_e_relationship_type(1)
EXPECTED_SECTIONS['tofc'].set_e_record_type(1)
EXPECTED_SECTIONS['tofc'].set_e_record_data_type(3)
EXPECTED_SECTIONS['tofc'].set_i_data_type_size(32)
EXPECTED_SECTIONS['tofc'].set_i_record_size(4)
EXPECTED_SECTIONS['tofc'].set_wc_data_unit('ns')
EXPECTED_SECTIONS['tofc'].set_accepted_units(['ns'])

EXPECTED_SECTIONS['tofb'] = AptFileSectionMetadata()
EXPECTED_SECTIONS['tofb'].set_section_name('tofb')
EXPECTED_SECTIONS['tofb'].set_c_signature()
EXPECTED_SECTIONS['tofb'].set_i_header_size(148)
EXPECTED_SECTIONS['tofb'].set_i_header_version(2)
EXPECTED_SECTIONS['tofb'].set_wc_section_type('tofb')
EXPECTED_SECTIONS['tofb'].set_i_section_version(1)
EXPECTED_SECTIONS['tofb'].set_e_relationship_type(1)
EXPECTED_SECTIONS['tofb'].set_e_record_type(1)
EXPECTED_SECTIONS['tofb'].set_e_record_data_type(3)
EXPECTED_SECTIONS['tofb'].set_i_data_type_size(32)
EXPECTED_SECTIONS['tofb'].set_i_record_size(4)
EXPECTED_SECTIONS['tofb'].set_wc_data_unit('ns')
EXPECTED_SECTIONS['tofb'].set_accepted_units(['ns'])

EXPECTED_SECTIONS['xs'] = AptFileSectionMetadata()
EXPECTED_SECTIONS['xs'].set_section_name('xs')
EXPECTED_SECTIONS['xs'].set_c_signature()
EXPECTED_SECTIONS['xs'].set_i_header_size(148)
EXPECTED_SECTIONS['xs'].set_i_header_version(2)
EXPECTED_SECTIONS['xs'].set_wc_section_type('xs')
EXPECTED_SECTIONS['xs'].set_i_section_version(1)
EXPECTED_SECTIONS['xs'].set_e_relationship_type(1)
EXPECTED_SECTIONS['xs'].set_e_record_type(1)
EXPECTED_SECTIONS['xs'].set_e_record_data_type(3)
EXPECTED_SECTIONS['xs'].set_i_data_type_size(32)
EXPECTED_SECTIONS['xs'].set_i_record_size(4)
EXPECTED_SECTIONS['xs'].set_wc_data_unit('nm')
EXPECTED_SECTIONS['xs'].set_accepted_units(['nm'])

EXPECTED_SECTIONS['ys'] = AptFileSectionMetadata()
EXPECTED_SECTIONS['ys'].set_section_name('ys')
EXPECTED_SECTIONS['ys'].set_c_signature()
EXPECTED_SECTIONS['ys'].set_i_header_size(148)
EXPECTED_SECTIONS['ys'].set_i_header_version(2)
EXPECTED_SECTIONS['ys'].set_wc_section_type('ys')
EXPECTED_SECTIONS['ys'].set_i_section_version(1)
EXPECTED_SECTIONS['ys'].set_e_relationship_type(1)
EXPECTED_SECTIONS['ys'].set_e_record_type(1)
EXPECTED_SECTIONS['ys'].set_e_record_data_type(3)
EXPECTED_SECTIONS['ys'].set_i_data_type_size(32)
EXPECTED_SECTIONS['ys'].set_i_record_size(4)
EXPECTED_SECTIONS['ys'].set_wc_data_unit('nm')
EXPECTED_SECTIONS['ys'].set_accepted_units(['nm'])

EXPECTED_SECTIONS['zs'] = AptFileSectionMetadata()
EXPECTED_SECTIONS['zs'].set_section_name('zs')
EXPECTED_SECTIONS['zs'].set_c_signature()
EXPECTED_SECTIONS['zs'].set_i_header_size(148)
EXPECTED_SECTIONS['zs'].set_i_header_version(2)
EXPECTED_SECTIONS['zs'].set_wc_section_type('zs')
EXPECTED_SECTIONS['zs'].set_i_section_version(1)
EXPECTED_SECTIONS['zs'].set_e_relationship_type(1)
EXPECTED_SECTIONS['zs'].set_e_record_type(1)
EXPECTED_SECTIONS['zs'].set_e_record_data_type(3)
EXPECTED_SECTIONS['zs'].set_i_data_type_size(32)
EXPECTED_SECTIONS['zs'].set_i_record_size(4)
EXPECTED_SECTIONS['zs'].set_wc_data_unit('nm')
EXPECTED_SECTIONS['zs'].set_accepted_units(['nm'])

EXPECTED_SECTIONS['rTip'] = AptFileSectionMetadata()
EXPECTED_SECTIONS['rTip'].set_section_name('rTip')
EXPECTED_SECTIONS['rTip'].set_c_signature()
EXPECTED_SECTIONS['rTip'].set_i_header_size(148)
EXPECTED_SECTIONS['rTip'].set_i_header_version(2)
EXPECTED_SECTIONS['rTip'].set_wc_section_type('rTip')
EXPECTED_SECTIONS['rTip'].set_i_section_version(1)
EXPECTED_SECTIONS['rTip'].set_e_relationship_type(1)
EXPECTED_SECTIONS['rTip'].set_e_record_type(1)
EXPECTED_SECTIONS['rTip'].set_e_record_data_type(3)
EXPECTED_SECTIONS['rTip'].set_i_data_type_size(32)
EXPECTED_SECTIONS['rTip'].set_i_record_size(4)
EXPECTED_SECTIONS['rTip'].set_wc_data_unit('nm')
EXPECTED_SECTIONS['rTip'].set_accepted_units(['nm'])

EXPECTED_SECTIONS['zApex'] = AptFileSectionMetadata()
EXPECTED_SECTIONS['zApex'].set_section_name('zApex')
EXPECTED_SECTIONS['zApex'].set_c_signature()
EXPECTED_SECTIONS['zApex'].set_i_header_size(148)
EXPECTED_SECTIONS['zApex'].set_i_header_version(2)
EXPECTED_SECTIONS['zApex'].set_wc_section_type('zApex')
EXPECTED_SECTIONS['zApex'].set_i_section_version(1)
EXPECTED_SECTIONS['zApex'].set_e_relationship_type(1)
EXPECTED_SECTIONS['zApex'].set_e_record_type(1)
EXPECTED_SECTIONS['zApex'].set_e_record_data_type(3)
EXPECTED_SECTIONS['zApex'].set_i_data_type_size(32)
EXPECTED_SECTIONS['zApex'].set_i_record_size(4)
EXPECTED_SECTIONS['zApex'].set_wc_data_unit('nm')
EXPECTED_SECTIONS['zApex'].set_accepted_units(['nm'])

EXPECTED_SECTIONS['zSphereCorr'] = AptFileSectionMetadata()
EXPECTED_SECTIONS['zSphereCorr'].set_section_name('zSphereCorr')
EXPECTED_SECTIONS['zSphereCorr'].set_c_signature()
EXPECTED_SECTIONS['zSphereCorr'].set_i_header_size(148)
EXPECTED_SECTIONS['zSphereCorr'].set_i_header_version(2)
EXPECTED_SECTIONS['zSphereCorr'].set_wc_section_type('zSphereCorr')
EXPECTED_SECTIONS['zSphereCorr'].set_i_section_version(1)
EXPECTED_SECTIONS['zSphereCorr'].set_e_relationship_type(1)
EXPECTED_SECTIONS['zSphereCorr'].set_e_record_type(1)
EXPECTED_SECTIONS['zSphereCorr'].set_e_record_data_type(3)
EXPECTED_SECTIONS['zSphereCorr'].set_i_data_type_size(32)
EXPECTED_SECTIONS['zSphereCorr'].set_i_record_size(4)
EXPECTED_SECTIONS['zSphereCorr'].set_wc_data_unit('nm')
EXPECTED_SECTIONS['zSphereCorr'].set_accepted_units(['nm'])

# the next three seem to (have been /were used for AMETEK development purposes
EXPECTED_SECTIONS['Position_0'] = AptFileSectionMetadata()
EXPECTED_SECTIONS['Position_0'].set_section_name('Position_0')
EXPECTED_SECTIONS['Position_0'].set_c_signature()
EXPECTED_SECTIONS['Position_0'].set_i_header_size(148 + 6 * 4)
EXPECTED_SECTIONS['Position_0'].set_i_header_version(2)
EXPECTED_SECTIONS['Position_0'].set_wc_section_type('Position_0')
EXPECTED_SECTIONS['Position_0'].set_i_section_version(1)
EXPECTED_SECTIONS['Position_0'].set_e_relationship_type(1)
EXPECTED_SECTIONS['Position_0'].set_e_record_type(1)
EXPECTED_SECTIONS['Position_0'].set_e_record_data_type(3)
EXPECTED_SECTIONS['Position_0'].set_i_data_type_size(32)
EXPECTED_SECTIONS['Position_0'].set_i_record_size(12)
EXPECTED_SECTIONS['Position_0'].set_wc_data_unit('')

EXPECTED_SECTIONS['Position_1'] = AptFileSectionMetadata()
EXPECTED_SECTIONS['Position_1'].set_section_name('Position_1')
EXPECTED_SECTIONS['Position_1'].set_c_signature()
EXPECTED_SECTIONS['Position_1'].set_i_header_size(148 + 6 * 4)
EXPECTED_SECTIONS['Position_1'].set_i_header_version(2)
EXPECTED_SECTIONS['Position_1'].set_wc_section_type('Position_1')
EXPECTED_SECTIONS['Position_1'].set_i_section_version(1)
EXPECTED_SECTIONS['Position_1'].set_e_relationship_type(1)
EXPECTED_SECTIONS['Position_1'].set_e_record_type(1)
EXPECTED_SECTIONS['Position_1'].set_e_record_data_type(3)
EXPECTED_SECTIONS['Position_1'].set_i_data_type_size(32)
EXPECTED_SECTIONS['Position_1'].set_i_record_size(12)
EXPECTED_SECTIONS['Position_1'].set_wc_data_unit('')

EXPECTED_SECTIONS['Position_2'] = AptFileSectionMetadata()
EXPECTED_SECTIONS['Position_2'].set_section_name('Position_2')
EXPECTED_SECTIONS['Position_2'].set_c_signature()
EXPECTED_SECTIONS['Position_2'].set_i_header_size(148 + 6 * 4)
EXPECTED_SECTIONS['Position_2'].set_i_header_version(2)
EXPECTED_SECTIONS['Position_2'].set_wc_section_type('Position_2')
EXPECTED_SECTIONS['Position_2'].set_i_section_version(1)
EXPECTED_SECTIONS['Position_2'].set_e_relationship_type(1)
EXPECTED_SECTIONS['Position_2'].set_e_record_type(1)
EXPECTED_SECTIONS['Position_2'].set_e_record_data_type(3)
EXPECTED_SECTIONS['Position_2'].set_i_data_type_size(32)
EXPECTED_SECTIONS['Position_2'].set_i_record_size(12)
EXPECTED_SECTIONS['Position_2'].set_wc_data_unit('')

EXPECTED_SECTIONS['XDet_mm'] = AptFileSectionMetadata()
EXPECTED_SECTIONS['XDet_mm'].set_section_name('XDet_mm')
EXPECTED_SECTIONS['XDet_mm'].set_c_signature()
EXPECTED_SECTIONS['XDet_mm'].set_i_header_size(148)
EXPECTED_SECTIONS['XDet_mm'].set_i_header_version(2)
EXPECTED_SECTIONS['XDet_mm'].set_wc_section_type('XDet_mm')
EXPECTED_SECTIONS['XDet_mm'].set_i_section_version(1)
EXPECTED_SECTIONS['XDet_mm'].set_e_relationship_type(1)
EXPECTED_SECTIONS['XDet_mm'].set_e_record_type(1)
EXPECTED_SECTIONS['XDet_mm'].set_e_record_data_type(3)
EXPECTED_SECTIONS['XDet_mm'].set_i_data_type_size(32)
EXPECTED_SECTIONS['XDet_mm'].set_i_record_size(4)
EXPECTED_SECTIONS['XDet_mm'].set_wc_data_unit('mm')
EXPECTED_SECTIONS['XDet_mm'].set_accepted_units(['mm'])

EXPECTED_SECTIONS['YDet_mm'] = AptFileSectionMetadata()
EXPECTED_SECTIONS['YDet_mm'].set_section_name('YDet_mm')
EXPECTED_SECTIONS['YDet_mm'].set_c_signature()
EXPECTED_SECTIONS['YDet_mm'].set_i_header_size(148)
EXPECTED_SECTIONS['YDet_mm'].set_i_header_version(2)
EXPECTED_SECTIONS['YDet_mm'].set_wc_section_type('YDet_mm')
EXPECTED_SECTIONS['YDet_mm'].set_i_section_version(1)
EXPECTED_SECTIONS['YDet_mm'].set_e_relationship_type(1)
EXPECTED_SECTIONS['YDet_mm'].set_e_record_type(1)
EXPECTED_SECTIONS['YDet_mm'].set_e_record_data_type(3)
EXPECTED_SECTIONS['YDet_mm'].set_i_data_type_size(32)
EXPECTED_SECTIONS['YDet_mm'].set_i_record_size(4)
EXPECTED_SECTIONS['YDet_mm'].set_wc_data_unit('mm')
EXPECTED_SECTIONS['YDet_mm'].set_accepted_units(['mm'])

EXPECTED_SECTIONS['Multiplicity'] = AptFileSectionMetadata()
EXPECTED_SECTIONS['Multiplicity'].set_section_name('Multiplicity')
EXPECTED_SECTIONS['Multiplicity'].set_c_signature()
EXPECTED_SECTIONS['Multiplicity'].set_i_header_size(148)
EXPECTED_SECTIONS['Multiplicity'].set_i_header_version(2)
EXPECTED_SECTIONS['Multiplicity'].set_wc_section_type('Multiplicity')
EXPECTED_SECTIONS['Multiplicity'].set_i_section_version(1)
EXPECTED_SECTIONS['Multiplicity'].set_e_relationship_type(1)
EXPECTED_SECTIONS['Multiplicity'].set_e_record_type(1)
EXPECTED_SECTIONS['Multiplicity'].set_e_record_data_type(1)
EXPECTED_SECTIONS['Multiplicity'].set_i_data_type_size(32)
EXPECTED_SECTIONS['Multiplicity'].set_i_record_size(4)
EXPECTED_SECTIONS['Multiplicity'].set_wc_data_unit('')

EXPECTED_SECTIONS['Vap'] = AptFileSectionMetadata()
EXPECTED_SECTIONS['Vap'].set_section_name('Vap')
EXPECTED_SECTIONS['Vap'].set_c_signature()
EXPECTED_SECTIONS['Vap'].set_i_header_size(148)
EXPECTED_SECTIONS['Vap'].set_i_header_version(2)
EXPECTED_SECTIONS['Vap'].set_wc_section_type('Vap')
EXPECTED_SECTIONS['Vap'].set_i_section_version(1)
EXPECTED_SECTIONS['Vap'].set_e_relationship_type(1)
EXPECTED_SECTIONS['Vap'].set_e_record_type(1)
EXPECTED_SECTIONS['Vap'].set_e_record_data_type(3)
EXPECTED_SECTIONS['Vap'].set_i_data_type_size(32)
EXPECTED_SECTIONS['Vap'].set_i_record_size(4)
EXPECTED_SECTIONS['Vap'].set_wc_data_unit('V')
EXPECTED_SECTIONS['Vap'].set_accepted_units(['V'])

EXPECTED_SECTIONS['Detector Coordinates'] = AptFileSectionMetadata()
EXPECTED_SECTIONS['Detector Coordinates'].set_section_name(
    'Detector Coordinates')
EXPECTED_SECTIONS['Detector Coordinates'].set_c_signature()
EXPECTED_SECTIONS['Detector Coordinates'].set_i_header_size(148)
EXPECTED_SECTIONS['Detector Coordinates'].set_i_header_version(2)
EXPECTED_SECTIONS['Detector Coordinates'].set_wc_section_type(
    'Detector Coordinates')
EXPECTED_SECTIONS['Detector Coordinates'].set_i_section_version(1)
EXPECTED_SECTIONS['Detector Coordinates'].set_e_relationship_type(1)
EXPECTED_SECTIONS['Detector Coordinates'].set_e_record_type(1)
EXPECTED_SECTIONS['Detector Coordinates'].set_e_record_data_type(3)
EXPECTED_SECTIONS['Detector Coordinates'].set_i_data_type_size(32)
EXPECTED_SECTIONS['Detector Coordinates'].set_i_record_size(8)
EXPECTED_SECTIONS['Detector Coordinates'].set_wc_data_unit('mm')
EXPECTED_SECTIONS['Detector Coordinates'].set_accepted_units(['mm'])

# there is at least a hint that some time ago this section's wcDataUnit was ''


# sections whose formatting and purpose is completely unclear

# Var44 ? magic-in-action? M. K\"uhbach, could well for AMETEK development only
# EXPECTED_SECTIONS['Var44'].set_section_name('Var44')


# deprecated sections or sections with detected inconsistencies across versions
# Vref vs Voltage branch issue
EXPECTED_SECTIONS['Vref'] = EXPECTED_SECTIONS['Voltage']
EXPECTED_SECTIONS['Vref'].set_wc_data_unit('V')
EXPECTED_SECTIONS['Vref'].set_accepted_units(['V'])

# pulseDelta vs Delta Pulse issue
# at least in one case a section Delta Pulse appeared
# at least in one case a section Epos ToF appeared

# other comments and issues
# Need to check APSuite version and build number
# 'Voltage' section, M. K\"uhbach: expect to have units but example *.apt files
# from flat test do not encode a unit in the 'Voltage' section



# Set of utility tools for parsing file formats used by atom probe.
#
# Also convenience functions are included which translate human-readable ion
# names into the isotope_vector description proposed by Kuehbach et al. in
# DOI: 10.1017/S1431927621012241 to the human-readable ion names which are use
# in P. Felfer et al.'s atom probe toolbox
#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

# pylint: disable=no-member,duplicate-code

import typing

from typing import Tuple

import numpy as np

from ifes_apt_tc_data_modeling.utils.definitions import \
    MAX_NUMBER_OF_ATOMS_PER_ION
from ifes_apt_tc_data_modeling.utils.utils import \
    create_isotope_vector, isotope_vector_to_nuclid_list, \
    isotope_vector_to_human_readable_name, \
    is_range_overlapping, is_range_significant

from ifes_apt_tc_data_modeling.utils.molecular_ions import MolecularIonCandidate

from ifes_apt_tc_data_modeling.nexus.nx_field import NxField


class NxIon():
    """Representative of a NeXus base class NXion."""

    def __init__(self, *args, **kwargs):
        self.comment = NxField("", "")  # comment, use e.g. for label of custom ion types
        self.color = NxField("", "")  # color used by software which created the dataset
        self.volume = NxField("", "")  # volume value in range files
        self.ion_type = NxField("", "")
        if len(args) >= 1:
            assert isinstance(args[0], list), "args[0] needs to be a list !"
            self.isotope_vector = NxField(create_isotope_vector(args[0]), "")
        elif "isotope_vector" in kwargs:
            assert isinstance(kwargs["isotope_vector"], np.ndarray), \
                "kwargs isotope_vector needs to be an np.ndarray !"
            assert np.shape(kwargs["isotope_vector"]) \
                == (MAX_NUMBER_OF_ATOMS_PER_ION,), \
                "kwargs isotope_vector needs be a " \
                + "(" + str(MAX_NUMBER_OF_ATOMS_PER_ION) + ",) array!"
            self.isotope_vector \
                = NxField(np.asarray(kwargs["isotope_vector"], np.uint16), "")
        else:
            # the default UNKNOWN IONTYPE
            self.isotope_vector = NxField(create_isotope_vector([]), "")
        self.nuclid_list = NxField(
            isotope_vector_to_nuclid_list(self.isotope_vector.typed_value), "")
        if "charge_state" in kwargs:
            if isinstance(kwargs["charge_state"], int) \
                    and (-8 < kwargs["charge_state"] < +8):
                self.charge_state = NxField(np.int8(kwargs["charge_state"]), "")
        else:
            # charge_state 0 flags and warns that it was impossible to recover
            # the relevant charge which is usually a sign that the range
            # is not matching the theoretically expect peak location
            self.charge_state = NxField(np.int8(0), "")
        self.name = NxField(isotope_vector_to_human_readable_name(
            self.isotope_vector.typed_value, self.charge_state.typed_value))
        self.ranges = NxField(np.empty((0, 2), np.float64), "amu")

    def add_range(self, mqmin: np.float64, mqmax: np.float64):
        """Adding mass-to-charge-state ratio interval."""
        assert is_range_significant(mqmin, mqmax) is True, \
            "Refusing to add epsilon range!"
        # the following example shows that is_range_overlapping should not be checked for
        # like it was in the past
        # ion.add_range(10.0, 12.0), ion.add_range(12.0, 13.3)
        # is equivalent to ion.add_range(10.0, 13.3)
        # assert is_range_overlapping(np.asarray([mqmin, mqmax]),
        #                            self.ranges.typed_value) is False, \
        #                            "Refusing overlapping range!"
        self.ranges.typed_value = np.vstack(
            (self.ranges.typed_value, np.array([mqmin, mqmax])))

    def update_human_readable_name(self):
        """Reevaluate charge and isotope_vector for name."""
        self.name = NxField(isotope_vector_to_human_readable_name(
            self.isotope_vector.typed_value, self.charge_state.typed_value))

    def report(self):
        """Report values."""
        print("ion_type")
        print(self.ion_type.typed_value)
        print("isotope_vector")
        print(self.isotope_vector.typed_value)
        print("nuclid_list")
        print(self.nuclid_list.typed_value)
        print("human-readable name")
        print(self.name.typed_value)
        print("charge_state")
        print(self.charge_state.typed_value)
        print("ranges")
        print(self.ranges.typed_value)
        print("comment")
        print(self.comment.typed_value)
        print("color")
        print(self.color.typed_value)
        print("volume")
        print(self.volume.typed_value)

    def add_charge_state_model(self,
                         parameters={},
                         candidates=[]):
        """Add details about the model how self.charge_state was defined."""
        self.charge_state_model = {}
        assert "min_abundance" in parameters.keys(), \
            "Parameter min_abundance not defined!"
        assert "min_abundance_product" in parameters.keys(), \
            "Parameter min_abundance_product not defined!"
        assert "min_half_life" in parameters.keys(), \
            "Parameter min_half_life not defined!"
        assert "sacrifice_isotopic_uniqueness" in parameters.keys(), \
            "Parameter sacrifice_isotopic_uniqueness not defined!"
        self.charge_state_model = {
            "isotope_matrix": [],
            "charge_state_vector": [],
            "mass_vector": [],
            "nat_abun_prod_vector": [],
            "min_half_life_vector": []}
        for key, val in parameters.items():
            if key not in self.charge_state_model.keys():
                self.charge_state_model[key] = val
        n_cand = len(candidates)
        # print("n_cand " + str(n_cand))
        if n_cand > 0:
            self.charge_state_model["isotope_matrix"] = np.zeros(
                (n_cand, MAX_NUMBER_OF_ATOMS_PER_ION), np.uint16)
            self.charge_state_model["charge_state_vector"] = np.zeros(
                (n_cand, ), np.int8)
            self.charge_state_model["mass_vector"] = np.zeros(
                (n_cand, ), np.float64)
            self.charge_state_model["nat_abun_prod_vector"] = np.zeros(
                (n_cand, ), np.float64)
            self.charge_state_model["min_half_life_vector"] = np.zeros(
                (n_cand, ), np.float64)
            row = 0
            for cand in candidates:
                if isinstance(cand, MolecularIonCandidate):
                    self.charge_state_model["isotope_matrix"][row, 0:len(cand.isotope_vector)] \
                        = cand.isotope_vector
                    self.charge_state_model["charge_state_vector"][row] = cand.charge_state
                    self.charge_state_model["mass_vector"][row] = cand.mass
                    self.charge_state_model["nat_abun_prod_vector"][row] \
                        = cand.abundance_product
                    self.charge_state_model["min_half_life_vector"][row] \
                        = cand.shortest_half_life
                    row += 1
                else:
                    print(__name__ + " found cand which is not a MolecularIonCandidate!")
        # else:
        #     print("Not enough candidates to report as a charge_state model")



# POS file format reader used by atom probe microscopists.
#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

# pylint: disable=no-member,duplicate-code

import os

import numpy as np

from ifes_apt_tc_data_modeling.nexus.nx_field import NxField

from ifes_apt_tc_data_modeling.utils.mmapped_io import get_memory_mapped_data


class ReadPosFileFormat():
    """Read *.pos file format."""

    def __init__(self, filename: str):
        assert len(filename) > 4, "POS file incorrect filename ending!"
        assert filename.lower().endswith(".pos"), \
            "POS file incorrect file type!"
        self.filename = filename

        self.filesize = os.path.getsize(self.filename)
        assert self.filesize % 4 * 4 == 0, \
            "POS filesize not integer multiple of 4*4B!"
        assert np.uint32(self.filesize / (4 * 4)) < np.iinfo(np.uint32).max, \
            "POS file is too large, currently only 2*32 supported!"
        self.number_of_events = np.uint32(self.filesize / (4 * 4))
        # print("Initialized access to " + self.filename + " successfully")

        # https://doi.org/10.1007/978-1-4614-3436-8 for file format details
        # dtyp_names = ["Reconstructed position along the x-axis (nm)",
        #               "Reconstructed position along the y-axis (nm)",
        #               "Reconstructed position along the z-axis (nm)",
        #               "Reconstructed mass-to-charge-state ratio (Da)"]

    def get_reconstructed_positions(self):
        """Read xyz columns."""

        xyz = NxField()
        xyz.typed_value = np.zeros(
            [self.number_of_events, 3], np.float32)
        xyz.unit = "nm"

        xyz.typed_value[:, 0] = \
            get_memory_mapped_data(self.filename, ">f4",
                                   0 * 4, 4 * 4, self.number_of_events)  # x
        xyz.typed_value[:, 1] = \
            get_memory_mapped_data(self.filename, ">f4",
                                   1 * 4, 4 * 4, self.number_of_events)  # y
        xyz.typed_value[:, 2] = \
            get_memory_mapped_data(self.filename, ">f4",
                                   2 * 4, 4 * 4, self.number_of_events)  # z
        return xyz

    def get_mass_to_charge_state_ratio(self):
        """Read mass-to-charge-state-ratio column."""

        m_n = NxField()
        m_n.typed_value = np.zeros(
            [self.number_of_events, 1], np.float32)
        m_n.unit = "Da"

        m_n.typed_value[:, 0] = \
            get_memory_mapped_data(self.filename, ">f4",
                                   3 * 4, 4 * 4, self.number_of_events)
        return m_n



# AMETEK APT(6) data exchange file reader used by atom probe microscopists.
#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

# pylint: disable=no-member,duplicate-code

import numpy as np

from ifes_apt_tc_data_modeling.apt.apt6_utils import string_to_typed_nparray


class AptFileHeaderMetadata():
    """Information content in the header to an APT(6) file."""

    # we make the variable names as close as possible to the original naming
    # scheme from the source code snippets which AMETEK shared with us
    # when developing the parser for their vendor file format
    def __init__(self):
        # file format signature
        self.c_signature = string_to_typed_nparray('APT\0', 4, np.uint8)
        # byte length of the file header
        self.i_header_size = np.int32(540)
        # version number of the file header, currently expecting 2
        self.i_header_version = np.int32(2)
        # original filename, i.e. *.apt filename, null-terminated UTF-16 !
        self.wc_filename = string_to_typed_nparray('', 256, np.uint16)
        # file creation time
        # according to AMETEK is implemented as a VisualStudio C++ FILETIME
        # 64-bit value, which represents the number of 100-nanosecond intervals
        # since January 1, 1601, as per the MSDN specification
        self.ft_creation_time = np.uint64(0)
        # number of ions represented by file
        self.ll_ion_count = np.uint64(0)  # or an int64 ?

    @classmethod
    def get_numpy_struct(cls) -> np.dtype:
        """Create customized numpy struct to read a file header at once."""
        return np.dtype([('cSignature', np.uint8, (4,)),
                         ('iHeaderSize', np.int32),
                         ('iHeaderVersion', np.int32),
                         ('wcFilename', np.uint16, 256),
                         ('ftCreationTime', np.uint64),
                         ('llIonCount', np.uint64)])

    def set_ll_ion_count(self, value: np.uint64):
        """Check and set total ion count."""
        assert isinstance(value, np.uint64), \
            'llIonCount needs to be an int!'
        assert value > 0, \
            'llIonCount needs to be positive and not zero!'
        assert value <= np.iinfo(np.uint64).max, \
            'llIonCount is too large, needs to map to np.uint64!'
        self.ll_ion_count = np.uint64(value)

    def matches(self, found_header: np.ndarray) -> bool:
        """Compare a read header against expectation."""
        assert np.array_equal(self.c_signature,
                              found_header['cSignature'][0],
                              equal_nan=True), \
            'Header cSignature differs!'
        assert self.i_header_size \
            == found_header['iHeaderSize'][0], \
            'Header iHeaderSize differs!'
        assert self.i_header_version \
            == found_header['iHeaderVersion'][0], \
            'Header iHeaderVersion differs!'
        assert found_header['llIonCount'][0] > 0, \
            'Header indicates there are no ions in the file!'

        self.set_ll_ion_count(found_header['llIonCount'][0])
        return True


# define which header and sections to expect in an *.apt file
# the sections below are referred to as branches in the commercial software
# APSuite enables users select prior I/O which of the sections to write
# normally a range file from based on the exchange with AMETEK and
# expected_header = apt6file_header_metadata()
EXPECTED_HEADER = AptFileHeaderMetadata()



# FIG.TXT range file reader used by atom probe microscopists.
#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

# pylint: disable=no-member,duplicate-code

import re

import numpy as np

from ifes_apt_tc_data_modeling.nexus.nx_ion import NxField, NxIon

from ifes_apt_tc_data_modeling.utils.utils import \
    create_isotope_vector, is_range_significant

from ifes_apt_tc_data_modeling.utils.definitions import \
    MQ_EPSILON, MAX_NUMBER_OF_ATOMS_PER_ION

from ifes_apt_tc_data_modeling.utils.molecular_ions import \
    isotope_to_hash, MolecularIonBuilder, \
    PRACTICAL_ABUNDANCE, PRACTICAL_ABUNDANCE_PRODUCT, \
    PRACTICAL_MIN_HALF_LIFE, VERBOSE, SACRIFICE_ISOTOPIC_UNIQUENESS

from ase.data import atomic_numbers, atomic_masses, chemical_symbols


class ReadFigTxtFileFormat():
    """Read *.fig.txt file format."""

    def __init__(self, filename: str):
        assert len(filename) > 7, "FIG.TXT file incorrect filename ending!"
        assert filename.lower().endswith(".fig.txt"), \
            "FIG.TXT file incorrect file type!"
        self.filename = filename

        self.fig: dict = {}
        self.fig["ranges"] = {}
        self.fig["ions"] = {}
        self.fig["molecular_ions"] = []
        self.read_fig_txt()

    def read_fig_txt(self):
        """Read FIG.TXT range file content."""
        with open(self.filename, mode="r", encoding="utf8") as figf:
            txt = figf.read()

        txt = txt.replace("\r\n", "\n")  # windows to unix EOL conversion
        txt = txt.replace(",", ".")  # use decimal dots instead of comma
        txt_stripped = [line for line in txt.split("\n")
                        if line.strip() != "" and line.startswith("#") is False]
        for molecular_ion in txt_stripped:
            tmp = molecular_ion.split(" ")
            mqmin = np.float64(tmp[len(tmp)-2:-1][0])
            mqmax = np.float64(tmp[len(tmp)-1:][0])
            ionname = " ".join(tmp[:-2])
            # print(f"{ionname} [{mqmin}, {mqmax}]")
            # ionname = '16O 1H2 + + +  + '

            positive = ionname.count("+")
            negative = ionname.count("-")
            if (0 < positive <= 7) and (negative == 0):
                charge_state = positive
            elif (0 < negative <= 7) and (positive == 0):
                charge_state = -negative
            else:
                charge_state = 0

            tmp = ionname.replace("+", "").replace("-", "").split(" ")
            ivec = []
            for isotope in tmp:
                if isotope != "":
                    prefix = re.findall("^[0-9]+", isotope)
                    mass_number = 0
                    if len(prefix) == 1:
                        if int(prefix[0]) > 0:
                            mass_number = int(prefix[0])
                    suffix = re.findall("[0-9]+$", isotope)
                    multiplier = 1
                    if len(suffix) == 1:
                        multiplier = int(suffix[0])
                    symbol = isotope.replace( \
                        str(mass_number), "").replace( \
                        str(multiplier), "").replace(" ", "")
                    if (symbol != "X") and (symbol in chemical_symbols):
                        for cnt in np.arange(0, multiplier):
                            proton_number = atomic_numbers[symbol]
                            neutron_number = mass_number - proton_number
                            ivec.append(isotope_to_hash(proton_number, neutron_number))
            ivec = np.sort(np.asarray(ivec, np.uint16))[::-1]
            ivector = np.zeros((MAX_NUMBER_OF_ATOMS_PER_ION,), np.uint16)
            ivector[0:len(ivec)] = ivec
            # print(ivector)

            m_ion = NxIon(isotope_vector=ivector, charge_state=charge_state)
            m_ion.add_range(mqmin, mqmax)
            m_ion.comment = NxField(ionname, "")
            # m_ion.report()

            crawler = MolecularIonBuilder(
                min_abundance=PRACTICAL_ABUNDANCE,
                min_abundance_product=PRACTICAL_ABUNDANCE_PRODUCT,
                min_half_life=PRACTICAL_MIN_HALF_LIFE,
                sacrifice_uniqueness=SACRIFICE_ISOTOPIC_UNIQUENESS,
                verbose=VERBOSE)
            recovered_charge_state, m_ion_candidates = crawler.combinatorics(
                m_ion.isotope_vector.typed_value,
                m_ion.ranges.typed_value[0, 0],
                m_ion.ranges.typed_value[0, 1])
            # print(f"{recovered_charge_state}")
            m_ion.charge_state = NxField(np.int8(recovered_charge_state), "")
            m_ion.update_human_readable_name()
            m_ion.add_charge_state_model(
                {"min_abundance": PRACTICAL_ABUNDANCE,
                 "min_abundance_product": PRACTICAL_ABUNDANCE_PRODUCT,
                 "min_half_life": PRACTICAL_MIN_HALF_LIFE,
                 "sacrifice_isotopic_uniqueness": SACRIFICE_ISOTOPIC_UNIQUENESS},
                m_ion_candidates)

            self.fig["molecular_ions"].append(m_ion)
        print(self.filename + " parsed successfully")

if __name__ == "main":
    pass
    # testing
    # parsedFile = ReadFigTxtFileFormat("../../02763-v01.rng.fig.txt")
